import meteva
import pandas as pd
import numpy as np
import math
import datetime


def AQI_value_to_grade(value):
    '''
    将AQI的具体值转换成AQI的等级
    :param value: AQI的具体数值
    :return: AQI的等级
    '''
    grade = np.abs(value)
    grade[value<=50] = 1
    index = np.where(value>50&value<=100)
    grade[index] = 2
    index = np.where(value>100&value<=150)
    grade[index] = 3
    index = np.where(value>150&value<=200)
    grade[index] = 4
    index = np.where(value>200&value<=300)
    grade[index] = 5
    grade[value>300] =6
    return grade

def O3_value_to_grade(value):
    '''
    将O3的具体值转换成O3的等级
    :param value: O3的具体数值
    :return: O3的等级
    '''
    grade = np.abs(value)
    grade[value<=160] = 1
    index = np.where(value>160&value<=200)
    grade[index] = 2
    index = np.where(value>200&value<=300)
    grade[index] = 3
    index = np.where(value>300&value<=400)
    grade[index] = 4
    index = np.where(value>400&value<=800)
    grade[index] = 5
    grade[value>800] =6
    return grade

def PM25_value_to_grade(value):
    '''
    将PM2.5的具体值转换成PM2.5的等级
    :param value: PM2.5的具体数值
    :return: PM2.5的等级
    '''
    grade = np.abs(value)
    grade[value<=35] = 1
    index = np.where(value>35&value<=75)
    grade[index] = 2
    index = np.where(value>75&value<=115)
    grade[index] = 3
    index = np.where(value>115&value<=150)
    grade[index] = 4
    index = np.where(value>150&value<=250)
    grade[index] = 5
    grade[value>250] =6
    return grade

def AQI_comprehensive(AQI_primary_type_score,AQI_grade_score,AQI_value_score,O3_grade_score,PM25_grade_score=None):
    '''
    AQI预报综合评分按如下公式进行计算：
    S1=0.2f1+0.2f2+0.2f3+0.2f4+0.2f5（全年）
    S1=0.2f1+0.4f2+0.2f3+0.2f4（5至10月）

    :param AQI_primary_type_score:首要污染物预报准确率评分
    :param AQI_grade_score:AQI级别预报准确率评分
    :param AQI_value_score:AQI级别预报准确率评分
    :param O3_grade_score:AQI级别预报准确率评分
    :param PM25_grade_score:PM2.5等级预报准确率评分
    :return:AQI预报综合评分（取1位小数）
    '''
    if PM25_grade_score is None:
        s1 = 0.25 * AQI_primary_type_score + 0.25 * AQI_grade_score + 0.25 * AQI_value_score + 0.25 * O3_grade_score
    else:
        s1 = 0.2 * AQI_primary_type_score + 0.4 * AQI_grade_score + 0.2 * AQI_value_score + 0.2 * O3_grade_score +0.2 * PM25_grade_score
    score = round(s1, 1)
    return score


def AQI_primary_type_score(ob_type,fo_type):
    '''
    计算首要污染物类型评分
    :param ob_type:  N×6的矩阵， N代表站点，考虑到首要污染物可能同时有多种，因此对于一个站点，用6个数字来表示首要污染物的类型
    6个数字的依次代表SO2,NO2,PM10,CO,O3,PM2.5是否为首要污染物，例如[0,1,0,0,1,0]代表首要污染物是PM10和O3。
    :param fo_type: 形式和观测相同
    :return: 首要污染物类型预报评分
    '''
    correct_rate = AQI_primary_type_correct_rate(ob_type,fo_type)
    score = correct_rate * 100
    return score


def grade_score(ob_grade,fo_grade):
    '''
    适用于计算AQI等级预报评分、O3等级预报评分和PM2.5等级预报评分
    :param ob_grade:  观测的等级
    :param fo_grade:  预报的等级
    :return: 等级预报评分
    '''
    grade_error = np.abs(ob_grade - fo_grade)
    score = np.zeros(grade_error.shape)
    score[grade_error==0] = 100
    score[grade_error==1] = 50
    score[grade_error==2] = 25
    score_mean = np.mean(score)
    return score_mean

def AQI_value_score(ob_value,fo_value):
    '''
    计算AQI 数值预报误差评分
    :param ob_value: AQI观测值
    :param fo_value: AQI预报值
    :return:
    '''

    error = np.abs(ob_value - fo_value)
    score = np.zeros(error.shape)
    score[error<=25] = 100
    index = np.where(error>25&error<=50)
    score[index] = 80
    index = np.where(error>50&error<=100)
    score[index] = 60
    index = np.where(error>100&error<=150)
    score[index] = 30
    score_mean = np.mean(score)
    return score_mean


def AQI_primary_type_correct_rate(ob_type,fo_type):
    '''
    计算首要污染物类型准确率
    :param ob_type:  N×6的矩阵， N代表站点，考虑到首要污染物可能同时有多种，因此对于一个站点，用6个数字来表示首要污染物的类型
    6个数字的依次代表SO2,NO2,PM10,CO,O3,PM2.5是否为首要污染物，例如[0,1,0,0,1,0]代表首要污染物是PM10和O3。
    :param fo_type: 形式和观测相同
    :return: 首要污染物类型准确率
    '''
    error = np.sum(np.abs(ob_type - fo_type))
    score = np.zeros(error.shape)
    score[error ==0] = 1 # 首要污染物完全一致是评分为1

    cross = ob_type * fo_type
    cross_sum = np.sum(cross,axis = -1)
    score[cross_sum>0] = 1  # 观测预报点积大于0，说明至少命中了1种
    score_mean = np.mean(score)
    return score_mean


def grade_correct_rate(ob_grade,fo_grade):
    '''
     适用于计算AQI等级准确率、O3等级准确率和PM2.5等级准确率
    :param ob_grade:  观测的等级
    :param fo_grade:  预报的等级
    :return: 等级准确率
    '''
    cr = meteva.method.correct_rate(ob_grade,fo_grade,grade_list=[0.5])
    return cr


def AQI_comprehensive_skill(AQI_primary_type_skill,AQI_grade_skill,O3_grade_skill,PM25_grade_skill=None):
    '''
    AQI预报综合评分技巧如下公式进行计算：
      S2=0.2*f6+0.4*f7+0.2*f8+0.2*f9（全年）
      S2=0.3*f6+0.4*f7+0.3*f8（5至10月）
    :param AQI_primary_type_skill:首要污染物预报准确率技巧
    :param AQI_grade_skill:AQI级别预报准确率技巧
    :param O3_grade_skill:O3级别预报准确率评分
    :param f9:PM2.5级别预报准确率技巧
    :return:
    '''
    if PM25_grade_skill is None:
        s2 = 0.3 * AQI_primary_type_skill + 0.4 * AQI_grade_skill + 0.3 * O3_grade_skill
    else:
        s2 = 0.2 * AQI_primary_type_skill + 0.4 * AQI_grade_skill + 0.2 * O3_grade_skill + 0.2 * PM25_grade_skill

    score = round(s2, 1)
    return score


def skill_of_province_to_nmc(correct_rate_nmc,correct_rate_province):
    '''
    适用于计算省台的AQI首要污染物类型准确率、AQI等级准确率、O3等级准确率或PM2.5等级准确率相对于中央台的技巧
    适用于单个站次上的技巧值，也适用于多个站次（多个时刻或多个站点）上平均的技巧值。
    :param correct_rate_nmc: 中央台的准确率
    :param correct_rate_province: 省台的准确率。
    :return:
    '''
    skill = correct_rate_province - correct_rate_nmc
    return skill


def mean_of_AQI_comprehensive_in_short_range(TTk24,TTk48,TTk72):
    '''

    :param score24:
    :param score48:
    :param score72:
    :return:
    '''
    TT = 0.5 * TTk24 + 0.3 * TTk48 + 0.2 * TTk72
    return TT




if __name__ == "__main__":
    import xarray as xr
    import pandas as pd
    path = r"\\10.20.73.112\hjhss\grid\MME\2022051820\AQI_sites_2564_2022051820.nc"
    dat = xr.open_dataset(path)
    print(dat)
    dtime = dat["time"].values
    lon = dat["sites_lon"].values
    lat = dat["sites_lat"].values
    AQI = dat["AQI"].values.flatten()
    CO = dat["CO"].values.flatten()
    NO2 = dat["NO2"].values.flatten()
    O3 = dat["O3"].values.flatten()
    PM10 = dat["PM10"].values.flatten()
    PM25 = dat["PM25"].values.flatten()
    SO2 = dat["SO2"].values.flatten()
    site = np.arange(len(lon))+1
    dtime_site = np.expand_dims(dtime,1).repeat(len(lon),axis=1).flatten()
    dtime_lon = np.expand_dims(lon,0).repeat(len(dtime),axis=0).flatten()
    dtime_lat = np.expand_dims(lat,0).repeat(len(dtime),axis=0).flatten()
    dtime_id = np.expand_dims(site,0).repeat(len(dtime),axis=0).flatten()
    df = pd.DataFrame({"dtime":dtime_site,"id":dtime_id,"lon":dtime_lon,"lat":dtime_lat,"AQI":AQI,
                       "CO":CO,"NO2":NO2,"O3":O3,"PM10":PM10,"PM25":PM25,"SO2":SO2})
    sta = meteva.base.sta_data(df)
    sta["time"] = datetime.datetime(2022,5,18,20)
    sta["level"] = 0

    sta_part = meteva.base.sele_by_para(sta,dtime = 24)
    meteva.base.scatter_sta(sta_part,save_path=r"H:\test_data\a.png",subplot="member")





import numpy as np
import matplotlib.pyplot as plt
import datetime
plt.rcParams['font.sans-serif']=['SimHei'] #用来正常显示中文标签\
plt.rcParams['axes.unicode_minus']=False #用来正常显示负号
#import seaborn as sns
import meteva
import matplotlib.patches as patches
import math
import pandas as pd
import json


def time_list_line_error(sta_ob_and_fos0,s = None,save_dir = None,save_path = None,show = False,dpi = 300,title = "多时效预报误差对比图",
                         sup_fontsize = 10,width = None,height = None,json_dir = None,json_path = None):
    sta_ob_and_fos1 = meteva.base.sele_by_dict(sta_ob_and_fos0, s)
    sta_ob_and_fos1 = meteva.base.sele_by_para(sta_ob_and_fos1,drop_IV=True)

    ids = list(set(sta_ob_and_fos1.loc[:,"id"]))
    ids.sort()
    nids = len(ids)

    if isinstance(title, list):
        if nids != len(title):
            print("手动设置的title数目和要绘制的图形数目不一致")
            return

    if save_path is not None:
        if isinstance(save_path,str):
            save_path = [save_path]
        if nids != len(save_path):
            print("手动设置的save_path数目和要绘制的图形数目不一致")
            return
    if json_path is not None:
        if isinstance(json_path,str):
            json_path = [json_path]
        if nids != len(json_path):
            print("手动设置的json_path数目和要绘制的图形数目不一致")
            return

    for n in range(nids):
        id = ids[n]
        sta_ob_and_fos = meteva.base.in_id_list(sta_ob_and_fos1,[id])
        times_fo = sta_ob_and_fos.loc[:, "time"].values
        times_fo = list(set(times_fo))
        # print(times_fo)
        if (len(times_fo) == 1):
            print("仅有单个起报时间的预报，程序退出")
            return
        times_fo.sort()
        times_fo = np.array(times_fo)
        dhs_fo = (times_fo[1:] - times_fo[0:-1])
        if isinstance(dhs_fo[0], np.timedelta64):
            dhs_fo = dhs_fo / np.timedelta64(1, 'h')
        else:
            dhs_fo = dhs_fo / datetime.timedelta(hours=1)
        dhs_fo_not0 = dhs_fo[dhs_fo != 0]
        dh_y = np.min(dhs_fo_not0)

        dhs = list(set(sta_ob_and_fos.loc[:, "dtime"].values))
        dhs.sort()
        dhs = np.array(dhs)
        # 以观预报时效间隔的最小单位
        ddhs = dhs[1:] - dhs[0:-1]
        dh_x = int(np.min(ddhs))

        if width is None:
            width = len(dhs) * 1.2
            if width > 8:width = 8
        if height is None:
            height = len(times_fo) * 1.0
            if height > 5:height = 5

        fig = plt.figure(figsize=(width, height),dpi = dpi)
        grid_plt = plt.GridSpec(len(times_fo), 1, hspace=0)

        time_f0 = times_fo[0]
        data_names = meteva.base.get_stadata_names(sta_ob_and_fos)
        error_array = np.zeros((len(sta_ob_and_fos.index),len(data_names)-1))
        for i in range(len(data_names)-1):
            error_array[:,i] = sta_ob_and_fos.values[:, 7+i] - sta_ob_and_fos.values[:, 6]
        vmax0 = np.max(error_array)
        vmin0 = np.min(error_array) - 0.1
        maxerr = np.maximum(vmax0,-vmin0)
        vmax = maxerr * 1.05
        vmin = -maxerr * 1.05
        #vmax = (vmax - vmin) * 1.2 + vmin
        dif = (vmax - vmin)/2
        inte = math.pow(10, math.floor(math.log10(dif)))
        # 用基本间隔，将最大最小值除于间隔后小数点部分去除，最后把间隔也整数化
        r = dif / inte
        if(r<1.5):
            inte = inte * 0.5
        elif r < 3 and r >= 1.5:
            inte = inte * 1
        elif r < 4.5 and r >= 3:
            inte = inte * 2
        elif r < 5.5 and r >= 4.5:
            inte = inte * 3
        elif r < 7 and r >= 5.5:
            inte = inte * 3
        elif r >= 7:
            inte = inte * 4
        yticks = np.array([-inte,0,inte])

        dtimes = sta_ob_and_fos["dtime"] * np.timedelta64(1, 'h')
        obtimes = sta_ob_and_fos['time'] + dtimes
        #obtimes[-1] = times_fo[0]
        time_all = list(set(obtimes))
        time_all.sort()
        #print(time_all)
        dtime_all = pd.Series(time_all) - times_fo[0]
        x_all = dtime_all/np.timedelta64(1, 'h')
        x_all = x_all.values
        x_plot, time_strs = meteva.product.program.get_x_ticks(time_all, width-1)
        x_plot += x_all[0]

        time_strs_null = []
        for i in range(len(time_strs)):
            time_strs_null.append("")

        all_y_label = []

        picture_ele_dict = {}
        picture_ele_dict["xticklabels"] = meteva.product.program.fun.get_time_str_list(time_all,row=3)
        picture_ele_dict["vmin"] = vmin
        picture_ele_dict["vmax"] = vmax
        picture_ele_dict["subplots"] ={}

        for i in range(len(times_fo)):
            ax = plt.subplot(grid_plt[i:i + 1, 0])
            picture_ele_dict["subplots"][i] = {}
            time_f1 = times_fo[-i - 1]
            dhour0 = (time_f1 - time_f0) / np.timedelta64(1, 'h')
            sta = meteva.base.in_time_list(sta_ob_and_fos, [time_f1])
            sta = sta.sort_values("dtime")
            x = dhour0 + sta.loc[:, "dtime"].values
            plt.plot(x,np.zeros(x.size),linewidth = sup_fontsize *0.07)
            for name in data_names[1:]:
                value = sta.loc[:, name].values - sta.iloc[:, 6].values
                #plt.plot(x, value, label=name,marker = ".",linewidth = sup_fontsize *0.1,markersize = sup_fontsize *0.3)

                #value = sta.loc[:, name].values
                index_iv = np.where(value == meteva.base.IV)
                dat0_all = meteva.base.tool.plot_tools.set_plot_IV_with_out_start_end(value)
                plt.plot(x, dat0_all, "--", linewidth=0.5, color="k")
                x_iv = x[index_iv[0]]
                dat0_iv = dat0_all[index_iv[0]]
                plt.plot(x_iv, dat0_iv, ".", color='k',markersize = sup_fontsize * 0.1)
                dat0_notiv = value.copy()
                dat0_notiv[dat0_notiv == meteva.base.IV] = np.nan
                plt.plot(x, dat0_notiv, label=name,marker = ".",linewidth = sup_fontsize * 0.1,markersize = sup_fontsize * 0.3)

                plt.ylim(vmin, vmax)
                plt.yticks(yticks,fontsize = sup_fontsize *0.6)
                plt.xlim(x_all[0],x_all[-1])
                plt.grid(linestyle='-.',linewidth = sup_fontsize *0.07)

                picture_ele_dict["subplots"][i][name] = {}
                picture_ele_dict["subplots"][i][name]["x"] = x.tolist()
                picture_ele_dict["subplots"][i][name]["value"] = value.tolist()


            time_f1 = meteva.base.tool.time_tools.all_type_time_to_datetime(time_f1)
            time_str = time_f1.strftime('%d{d}%H{h}').format(d='日', h='时')+"        "
            all_y_label.append(time_str)
            plt.ylabel(time_str, rotation='horizontal',fontsize = sup_fontsize * 0.75)
            if i ==0:
                plt.legend(loc="upper left", ncol=len(data_names),fontsize = sup_fontsize * 0.9)
                s1 = s
                if s1 is None:
                    s1 = {}
                    s1["id"] = id

                if isinstance(title,list):
                    title1 = title[n]
                else:
                    title1 = meteva.product.program.get_title_from_dict(title, s1, None, None,
                                                                    None)

                    title1 = title1.replace("\n","")
                plt.title(title1,fontsize = sup_fontsize)
                picture_ele_dict["title"] = title1

            #plt.hlines(0,x_plot[0],x_plot[-1],"k",linewidth = 0.5)
            if i == len(times_fo) - 1:
                plt.xticks(x_plot, time_strs,fontsize =  sup_fontsize * 0.8)
                plt.xlabel("实况时间",fontsize = sup_fontsize * 0.9)
            else:
                plt.xticks(x_plot,time_strs_null)

        picture_ele_dict["y_label"] = all_y_label

        rect_ylabel = [0.03, 0.45, 0.0, 0.0]  # 左下宽高
        ax_ylabel = plt.axes(rect_ylabel)
        ax_ylabel.axes.set_axis_off()
        plt.text(0, 0, "起报时间", fontsize=sup_fontsize * 0.9, rotation=90)

        save_path1 = None
        if save_path is None:
            if save_dir is None:
                show = True
            else:
                save_path1 = save_dir+"/" + str(id) + ".png"
        else:
            save_path1 = save_path[n]

        if save_path1 is not None:
            meteva.base.tool.path_tools.creat_path(save_path1)
            plt.savefig(save_path1,bbox_inches='tight')
            print("图片已保存至" + save_path1)
        if show:
            plt.show()
        plt.close()


        json_path1 = None
        if json_path is None:
            if json_dir is None:
                pass
            else:
                json_path1 = json_dir+"/" + str(id) + ".json"
        else:
            json_path1 = json_path[n]

        if json_path1 is not None:
            meteva.base.tool.path_tools.creat_path(json_path1)
            file = open(json_path1,"w")
            json.dump(picture_ele_dict,file)
            print("have printed pictrue elements to " + json_path1)


def time_list_line(sta_ob_and_fos0,s = None,save_dir = None,save_path = None,show = False,dpi = 300,title = "预报准确性和稳定性对比图",
                   sup_fontsize = 10,width = None,height = None,json_dir = None,json_path = None):
    sta_ob_and_fos1 = meteva.base.sele_by_dict(sta_ob_and_fos0, s)
    ids = list(set(sta_ob_and_fos1.loc[:,"id"]))
    ids.sort()
    nids = len(ids)
    if isinstance(title, list):
        if nids != len(title):
            print("手动设置的title数目和要绘制的图形数目不一致")
            return

    if save_path is not None:
        if isinstance(save_path,str):
            save_path = [save_path]
        if nids != len(save_path):
            print("手动设置的save_path数目和要绘制的图形数目不一致")
            return

    for n in range(nids):
        id = ids[n]
        #print(id)
        sta_ob_and_fos = meteva.base.in_id_list(sta_ob_and_fos1,[id])
        times_fo = sta_ob_and_fos.loc[:, "time"].values
        times_fo = list(set(times_fo))
        if (len(times_fo) == 1):
            print("仅有单个起报时间的预报，程序退出")
            return
        times_fo.sort()
        times_fo = np.array(times_fo)
        dhs_fo = (times_fo[1:] - times_fo[0:-1])
        if isinstance(dhs_fo[0], np.timedelta64):
            dhs_fo = dhs_fo / np.timedelta64(1, 'h')
        else:
            dhs_fo = dhs_fo / datetime.timedelta(hours=1)
        dhs_fo_not0 = dhs_fo[dhs_fo != 0]
        dh_y = np.min(dhs_fo_not0)

        dhs = list(set(sta_ob_and_fos.loc[:, "dtime"].values))
        dhs.sort()
        dhs = np.array(dhs)
        # 以观预报时效间隔的最小单位
        ddhs = dhs[1:] - dhs[0:-1]
        dh_x = int(np.min(ddhs))

        if width is None:
            width = len(dhs) * 1.2
            if width > 8:width = 8
            if width < 4:width = 4

        if height is None:
            height = len(times_fo) * 1
            if height > 5:height = 5

        fig = plt.figure(figsize=(width, height),dpi = dpi)
        grid_plt = plt.GridSpec(len(times_fo), 1, hspace=0)

        time_f0 = times_fo[0]
        data_names = meteva.base.get_stadata_names(sta_ob_and_fos)
        values = sta_ob_and_fos.iloc[:, 6:].values.flatten()
        values = values[values != meteva.base.IV]
        vmax = np.max(values)
        vmin = np.min(values) - 0.1
        vmax = (vmax - vmin) * 1.2 + vmin


        dtimes = sta_ob_and_fos["dtime"] * np.timedelta64(1, 'h')
        obtimes = sta_ob_and_fos['time'] + dtimes
        #obtimes[-1] = times_fo[0]
        time_all = list(set(obtimes))
        time_all.sort()
        #print(time_all)
        dtime_all = pd.Series(time_all) - times_fo[0]
        x_all = dtime_all/np.timedelta64(1, 'h')
        x_all = x_all.values
        #print(x_all)
        x_plot,time_strs = meteva.product.program.get_x_ticks(time_all,width-1)
        x_plot += x_all[0]

        time_strs_null = []
        for i in range(len(time_strs)):
            time_strs_null.append("")

        all_y_label = []
        picture_ele_dict = {}
        picture_ele_dict["xticklabels"] = meteva.product.program.fun.get_time_str_list(time_all,row=3)
        picture_ele_dict["vmin"] = vmin
        picture_ele_dict["vmax"] = vmax
        picture_ele_dict["subplots"] ={}

        for i in range(len(times_fo)):
            ax = plt.subplot(grid_plt[i:i + 1, 0])
            picture_ele_dict["subplots"][i] = {}
            time_f1 = times_fo[-i - 1]
            dhour0 = (time_f1 - time_f0) / np.timedelta64(1, 'h')
            sta = meteva.base.in_time_list(sta_ob_and_fos, [time_f1])
            sta = sta.sort_values("dtime")
            x = dhour0 + sta.loc[:, "dtime"].values
            for name in data_names:
                value = sta.loc[:, name].values
                index_iv = np.where(value == meteva.base.IV)
                dat0_all = meteva.base.tool.plot_tools.set_plot_IV_with_out_start_end(value)
                plt.plot(x, dat0_all, "--", linewidth=0.5, color="k")
                x_iv = x[index_iv[0]]
                dat0_iv = dat0_all[index_iv[0]]
                plt.plot(x_iv, dat0_iv, ".", color='k',markersize = sup_fontsize * 0.1)
                dat0_notiv = value.copy()
                dat0_notiv[dat0_notiv == meteva.base.IV] = np.nan
                plt.plot(x, dat0_notiv, label=name,marker = ".",linewidth = sup_fontsize * 0.1,markersize = sup_fontsize * 0.3)


                plt.ylim(vmin, vmax)
                plt.yticks(fontsize = sup_fontsize * 0.6)
                plt.xlim(x_all[0],x_all[-1])
                plt.grid(linestyle='-.')

                picture_ele_dict["subplots"][i][name] = {}
                picture_ele_dict["subplots"][i][name]["x"] = x.tolist()
                picture_ele_dict["subplots"][i][name]["value"] = value.tolist()

            time_f1 = meteva.base.tool.time_tools.all_type_time_to_datetime(time_f1)
            time_str = time_f1.strftime('%d{d}%H{h}').format(d='日', h='时')+"        "
            all_y_label.append(time_str)
            plt.ylabel(time_str, rotation='horizontal',fontsize = sup_fontsize *0.75)
            if i ==0:
                plt.legend(loc="upper left", ncol=len(data_names),fontsize = sup_fontsize *0.9)
                s1 = s
                if s1 is None:
                    s1 = {}
                s1["id"] = id

                if isinstance(title, list):
                    title1 = title[n]
                else:
                    title1 = meteva.product.program.get_title_from_dict(title, s1, None, None,
                                                                        None)
                    title1 = title1.replace("\n", "")
                plt.title(title1, fontsize=sup_fontsize)
                picture_ele_dict["title"] = title1

            if i == len(times_fo) - 1:
                #print(x_plot)
                plt.xticks(x_plot, time_strs,fontsize = sup_fontsize * 0.8)
                plt.xlabel("实况时间",fontsize = sup_fontsize * 0.9)
            else:
                plt.xticks(x_plot,time_strs_null)

        picture_ele_dict["y_label"] = all_y_label
        rect_ylabel = [0.03, 0.45, 0.0, 0.0]  # 左下宽高
        ax_ylabel = plt.axes(rect_ylabel)
        ax_ylabel.axes.set_axis_off()
        plt.text(0, 0, "起报时间", fontsize=sup_fontsize * 0.9, rotation=90)

        save_path1 = None
        if save_path is None:
            if save_dir is None:
                show = True
            else:
                save_path1 = save_dir+"/" + str(id) + ".png"
        else:
            save_path1 = save_path[n]
        if save_path1 is not None:
            meteva.base.tool.path_tools.creat_path(save_path1)
            plt.savefig(save_path1,bbox_inches='tight')
            print("图片已保存至" + save_path1)
        if show:
            plt.show()
        plt.close()
        json_path1 = None
        if json_path is None:
            if json_dir is None:
                pass
            else:
                json_path1 = json_dir+"/" + str(id) + ".json"
        else:
            json_path1 = json_path[n]

        if json_path1 is not None:
            meteva.base.tool.path_tools.creat_path(json_path1)
            file = open(json_path1,"w")
            json.dump(picture_ele_dict,file)
            print("have printed pictrue elements to " + json_path1)



def time_list_mesh_error(sta_ob_and_fos0,s = None,save_dir = None,save_path = None,
                   max_error = None,cmap_error = None,show = False,xtimetype = "mid",dpi = 300,annot =0,title = "多时效预报误差对比图",
                         sup_fontsize = 10,width = None,height = None,json_dir = None,json_path = None):
    '''

    :param sta_ob_and_fos0:
    :param s:
    :param save_dir:
    :param save_path:
    :param clev:
    :param cmap:
    :param plot_error:
    :param cmap_error:
    :param show:
    :param title:
    :return:
    '''

    if max_error is None:
        sta_ob_fos0_noIV = meteva.base.not_IV(sta_ob_and_fos0)
        values = sta_ob_fos0_noIV.values[:,6:].T
        dvalues = values[1:,:] - values[0,:]
        maxd = np.max(np.abs(dvalues))
    else:
        maxd = max_error

    sta_ob_and_fos1 = meteva.base.sele_by_dict(sta_ob_and_fos0, s)
    sta_ob_and_fos1 = meteva.base.sele_by_para(sta_ob_and_fos1,drop_IV=True)
    if(len(sta_ob_and_fos1.index) == 0):
        print("there is no data to verify")
        return
    ids = list(set(sta_ob_and_fos1.loc[:, "id"]))
    ids.sort()
    data_names = meteva.base.get_stadata_names(sta_ob_and_fos1)
    times_fo = sta_ob_and_fos1.loc[:, "time"].values
    times_fo = list(set(times_fo))
    if (len(times_fo) == 1):
        print("仅有单个起报时间的预报，程序退出")
        return
    times_fo.sort()
    times_fo = np.array(times_fo)
    #print(times_fo)

    dhs_fo = (times_fo[1:] - times_fo[0:-1])
    if isinstance(dhs_fo[0], np.timedelta64):
        dhs_fo = dhs_fo / np.timedelta64(1, 'h')
    else:
        dhs_fo = dhs_fo / datetime.timedelta(hours=1)
    dhs_fo_not0 = dhs_fo[dhs_fo != 0]
    dh_y = np.min(dhs_fo_not0)
    min_dtime = int(np.min(sta_ob_and_fos1["dtime"]))


    ob_time_s = sta_ob_and_fos1["time"] + sta_ob_and_fos1["dtime"] * np.timedelta64(1, 'h')
    times_ob = list(set(ob_time_s.values))
    times_ob.sort()
    times_ob = np.array(times_ob)

    dhs_ob = (times_ob[1:] - times_ob[0:-1])
    if isinstance(dhs_ob[0], np.timedelta64):
        dhs_ob = dhs_ob / np.timedelta64(1, 'h')
    else:
        dhs_ob = dhs_ob / datetime.timedelta(hours=1)

    dhs_ob_not0 = dhs_ob[dhs_ob != 0]
    dh_x = np.min(dhs_ob_not0)
    #print(dh_x)
    np.sum(dhs_fo_not0)
    row = int(np.sum(dhs_fo_not0)/dh_y)+1
    col = int(np.sum(dhs_ob_not0)/dh_x)+1
    #print(row)
    t_ob = []
    for t in times_ob:
        t_ob.append(meteva.base.all_type_time_to_datetime(t))

    y_ticks = []
    t_fo0= meteva.base.all_type_time_to_datetime(times_fo[0])
    step = int(math.ceil(row / 40))

    if step !=1 :
        while step * dh_y % 3 !=0:
            step +=1

    y_plot = np.arange(0,row,step)+0.
    yticklabels = []
    for j in range(0,row,1):
        jr = row - j - 1
        time_fo = t_fo0 + datetime.timedelta(hours=1) * dh_y * jr
        hour = time_fo.hour
        day = time_fo.day
        str1 = str(day) + "日" + str(hour) + "时"
        yticklabels.append(str1)
        if j%step ==0:
            y_ticks.append(str1)

    if width is None:
        width = 8

    x_plot,x_ticks = meteva.product.get_x_ticks(times_ob,width-2)
    x_plot /= dh_x
    #y_plot, y_ticks = meteva.product.get_y_ticks(times_fo, height)
    if xtimetype == "right":
        x_plot  = x_plot+0.5
    elif xtimetype == "left":
        x_plot = x_plot -0.5
    else:
        x_plot = x_plot

    if col >=120:
        annot = None
    # if annot:
    #     annot = col <120

    annot_size = width * 50 / col
    if annot_size >16:
        annot_size= 16

    nids = len(ids)
    nfo = len(data_names) - 1
    if isinstance(title, list):
        if nids * nfo != len(title):
            print("手动设置的title数目和要绘制的图形数目不一致")
            return

    if save_path is not None:
        if isinstance(save_path,str):
            save_path = [save_path]
        if nids * nfo != len(save_path):
            print("手动设置的save_path数目和要绘制的图形数目不一致")
            return
    kk = 0
    for d in range(nfo):
        data_name = data_names[d+1]
        sta_one_member = meteva.base.in_member_list(sta_ob_and_fos1, [data_names[0],data_name])
        #meteva.base.set_stadata_names(sta_ob_part2, [data_name])
        #sta_one_member = meteva.base.combine_join(sta_ob_part2, sta_fo_all2)
        #以最近的预报作为窗口中间的时刻
        for id in ids:
            picture_ele_dict = {}
            picture_ele_dict["xticklabels"] = meteva.product.program.fun.get_time_str_list(times_ob, row=3)
            picture_ele_dict["yticklabels"] = yticklabels
            
            sta_one_id = meteva.base.in_id_list(sta_one_member,id)
            dat = np.ones((col, row)) * meteva.base.IV
            for j in range(row):
                jr = row - j - 1
                time_fo = times_fo[0] + np.timedelta64(1, 'h') * dh_y * jr
                sta_on_row = meteva.base.in_time_list(sta_one_id,time_fo)
                dhx0 = (time_fo - times_ob[0])/np.timedelta64(1, 'h')
                dhxs = sta_on_row["dtime"].values + dhx0
                index_i = (dhxs/dh_x).astype(np.int16)
                dat[index_i,j] = sta_on_row.values[:,-1] - sta_on_row.values[:,-2]
            mask = np.zeros_like(dat.T)
            mask[dat.T == meteva.base.IV] = True
            picture_ele_dict["error"] = dat.tolist()

            if height is None:
                height = width * row / col + 2
            f, ax2 = plt.subplots(figsize=(width, height), nrows=1, edgecolor='black',dpi = dpi)
            plt.subplots_adjust(left=0.1, bottom=0.15, right=0.98, top=0.90)


            if cmap_error is None:
                cmap_error =meteva.base.cmaps.me_bwr

            #sns.heatmap(dat.T, ax=ax2, mask=mask, cmap=cmap_part, vmin=-maxd, vmax=maxd, center=None, robust=False, annot=annot,fmt='.0f'
            #, annot_kws = {'size': annot_size})

            cmap_part,clevs = meteva.base.color_tools.def_cmap_clevs(cmap_error,vmin = -maxd,vmax = maxd)
            meteva.base.tool.myheatmap(ax2, dat.T, cmap = cmap_part,clevs=clevs,  annot=annot, fontsize=sup_fontsize * 0.8)

            ax2.set_xlabel('实况时间',fontsize = sup_fontsize *  0.9)
            ax2.set_ylabel('起报时间',fontsize = sup_fontsize * 0.9)
            ax2.set_xticks(x_plot)
            ax2.set_xticklabels(x_ticks,rotation=360, fontsize=sup_fontsize * 0.8)
            ax2.set_yticks(y_plot)
            ax2.set_yticklabels(y_ticks, rotation=360, fontsize=sup_fontsize * 0.8)
            
            ax2.grid(linestyle='--', linewidth=0.5)
            ax2.set_ylim(row-0.5, -0.5)
            s1 = s
            if s1 is None:
                s1 = {}
                s1["id"] = id
                s1["member"] =[data_name]
            if isinstance(title,list):
                title1 = title[kk]
            else:
                if id in meteva.base.station_id_name_dict.keys():
                    title1 = title + "(" + data_name + ")" + "{\'id\':" + str(id) +meteva.base.station_id_name_dict[id] +"}"
                else:
                    title1 = title + "(" + data_name + ")" + "{\'id\':" + str(id) +  "}"

            ax2.set_title(title1, loc='left', fontweight='bold', fontsize=sup_fontsize)
            rect = patches.Rectangle((-0.5,-0.5), col, row, linewidth=0.8, edgecolor='k', facecolor='none')
            ax2.add_patch(rect)
            #plt.tick_params(top='on', right='on', which='both')  # 显示上侧和右侧的刻度
            plt.rcParams['xtick.direction'] = 'in'  # 将x轴的刻度线方向设置抄向内
            plt.rcParams['ytick.direction'] = 'in'  # 将y轴的刻度方知向设置向内

            save_path1 = None
            if(save_path is None):
                if save_dir is None:
                    show = True
                else:
                    save_path1 = save_dir +"/" +data_name+"_"+str(id) + ".png"
            else:
                save_path1 = save_path[kk]
            if save_path1 is not None:

                meteva.base.tool.path_tools.creat_path(save_path1)
                plt.savefig(save_path1,bbox_inches='tight')
                print("图片已保存至"+save_path1)
            if show:
                plt.show()
            plt.close()

            json_path1 = None
            if json_path is None:
                if json_dir is None:
                    pass
                else:
                    json_path1 = json_dir + "/" + data_name + "_" + str(id) + ".json"
            else:
                json_path1 = json_path[kk]

            if json_path1 is not None:
                meteva.base.tool.path_tools.creat_path(json_path1)
                file = open(json_path1, "w")
                json.dump(picture_ele_dict, file)
                print("have printed pictrue elements to " + json_path1)
            kk += 1
    return

def time_list_mesh(sta_ob_and_fos0,s = None,save_dir = None,save_path = None,
                   clev = None,cmap = None,plot_error = True,max_error = None,cmap_error= None,
                   show = False,xtimetype = "mid",dpi = 300,annot =0,title = "预报准确性和稳定性对比图",
                   sup_fontsize = 10,width = None,height = None,json_dir = None,json_path = None):
    '''

    :param sta_ob_and_fos0:
    :param s:
    :param save_dir:
    :param save_path:
    :param clev:
    :param cmap:
    :param plot_error:
    :param cmap_error:
    :param show:
    :param title:
    :return:
    '''

    if max_error is None:
        sta_ob_fos0_noIV = meteva.base.not_IV(sta_ob_and_fos0)
        values = sta_ob_fos0_noIV.values[:,6:].T
        if(values.size ==0):
            print("无有效的观测数据")
            return
        dvalues = values[1:,:] - values[0,:]
        maxd = np.max(np.abs(dvalues))
    else:
        maxd = max_error


    sta_ob_and_fos1 = meteva.base.sele_by_dict(sta_ob_and_fos0, s)
    ids = list(set(sta_ob_and_fos1.loc[:,"id"]))
    ids.sort()
    data_names = meteva.base.get_stadata_names(sta_ob_and_fos1)
    sta_ob_all1 = meteva.base.sele_by_para(sta_ob_and_fos1,member=[data_names[0]])
    sta_fo_all1 = meteva.base.sele_by_para(sta_ob_and_fos1, member=data_names[1:])
    times_fo = sta_fo_all1.loc[:, "time"].values
    times_fo = list(set(times_fo))
    if (len(times_fo) == 1):
        print("仅有单个起报时间的预报，程序退出")
        return
    times_fo.sort()
    times_fo = np.array(times_fo)
    #print(times_fo)

    dhs_fo = (times_fo[1:] - times_fo[0:-1])
    if isinstance(dhs_fo[0], np.timedelta64):
        dhs_fo = dhs_fo / np.timedelta64(1, 'h')
    else:
        dhs_fo = dhs_fo / datetime.timedelta(hours=1)
    dhs_fo_not0 = dhs_fo[dhs_fo != 0]
    dh_y = np.min(dhs_fo_not0)
    min_dtime = int(np.min(sta_fo_all1["dtime"]))
    sta_ob_part1 = meteva.base.between_dtime_range(sta_ob_all1,min_dtime,min_dtime+dh_y-0.1)
    sta_ob_part2 = meteva.base.move_fo_time(sta_ob_part1,dh_y)

    ob_time_s = sta_fo_all1["time"] + sta_fo_all1["dtime"] * np.timedelta64(1, 'h')
    times_ob = list(set(ob_time_s.values))
    times_ob.sort()
    times_ob = np.array(times_ob)

    dhs_ob = (times_ob[1:] - times_ob[0:-1])
    if isinstance(dhs_ob[0], np.timedelta64):
        dhs_ob = dhs_ob / np.timedelta64(1, 'h')
    else:
        dhs_ob = dhs_ob / datetime.timedelta(hours=1)
    dhs_ob_not0 = dhs_ob[dhs_ob != 0]
    dh_x = np.min(dhs_ob_not0)
    #print(dh_x)
    np.sum(dhs_fo_not0)
    row = int(np.sum(dhs_fo_not0)/dh_y)+2
    col = int(np.sum(dhs_ob_not0)/dh_x)+1
    #print(row)
    t_ob = []

    for t in times_ob:
        t_ob.append(meteva.base.all_type_time_to_datetime(t))

    #t_fo =[]
    #for t in times_fo:
    #    t_fo.append(meteva.base.all_type_time_to_datetime(t))


    y_ticks = []
    t_fo0= meteva.base.all_type_time_to_datetime(times_fo[0])
    step = int(math.ceil(row / 40))

    if step !=1 :
        while step * dh_y % 3 !=0:
            step +=1

    y_plot = np.arange(0,row,step)
    yticklabels = []
    for j in range(0,row,1):
        jr = row - j - 1
        time_fo = t_fo0 + datetime.timedelta(hours=1) * dh_y * jr
        hour = time_fo.hour
        day = time_fo.day
        str1 = str(day) + "日" + str(hour) + "时"
        yticklabels.append(str1)
        if j%step ==0:
            y_ticks.append(str1)


    if width is None:
        width = 8
    x_plot,x_ticks = meteva.product.get_x_ticks(times_ob,width-2)
    x_plot /= dh_x
    #y_plot, y_ticks = meteva.product.get_y_ticks(times_fo, height)
    if xtimetype == "right":
        x_plot  = x_plot+0.5
    elif xtimetype == "left":
        x_plot = x_plot -0.5

    if col >=120:
        annot = None
    # if annot:
    #     annot = col <120
    annot_size = width * 50 / col
    if annot_size >16:
        annot_size= 16

    nids = len(ids)
    nfo = len(data_names) - 1
    if isinstance(title, list):
        if plot_error:
            if 2 * nids * nfo != len(title):
                print("手动设置的title数目和要绘制的图形数目不一致")
                return
        else:
            if nids * nfo != len(title):
                print("手动设置的title数目和要绘制的图形数目不一致")
                return

    if save_path is not None:
        if isinstance(save_path,str):
            save_path = [save_path]
        if nids * nfo != len(save_path):
            print("手动设置的save_path数目和要绘制的图形数目不一致")
            return
    kk1 = 0
    kk2 = 0
    for d in range(len(data_names)-1):
        data_name = data_names[d+1]
        sta_fo_all2 = meteva.base.in_member_list(sta_fo_all1, data_name)
        meteva.base.set_stadata_names(sta_ob_part2, [data_name])
        sta_one_member = meteva.base.combine_join(sta_ob_part2, sta_fo_all2)
        #以最近的预报作为窗口中间的时刻

        for id in ids:
            picture_ele_dict = {}
            picture_ele_dict["xticklabels"] = meteva.product.program.fun.get_time_str_list(times_ob, row=3)
            picture_ele_dict["yticklabels"] = yticklabels

            sta_one_id = meteva.base.in_id_list(sta_one_member,id)
            dat = np.ones((col, row)) * meteva.base.IV
            for j in range(row):
                jr = row - j - 1
                time_fo = times_fo[0] + np.timedelta64(1, 'h') * dh_y * jr
                sta_on_row = meteva.base.in_time_list(sta_one_id,time_fo)
                dhx0 = (time_fo - times_ob[0])/np.timedelta64(1, 'h')
                dhxs = sta_on_row["dtime"].values + dhx0
                index_i = (dhxs/dh_x).astype(np.int16)
                dat[index_i,j] = sta_on_row.values[:,-1]
            mask = np.zeros_like(dat.T)
            mask[dat.T == meteva.base.IV] = True
            picture_ele_dict["dat"] = dat.tolist()
            vmin = np.min(dat[dat != meteva.base.IV])
            vmax = np.max(dat[dat != meteva.base.IV])
            #print(vmax)
            if plot_error:
                if height is None:
                    height = (width * row / col + 2) * 2
                f, (ax1, ax2)  = plt.subplots(figsize=(width, height),nrows = 2,edgecolor='black',dpi = dpi)
                plt.subplots_adjust(left=0.1, bottom=0.15, right=0.98, top=0.90,hspace=0.3)
                dvalue = np.zeros_like(dat)
                for i in range(col):
                    top_value = meteva.base.IV
                    for j in range(row):
                        if dat[i, j] != meteva.base.IV:
                            top_value = dat[i, j]
                            break
                    for j in range(row):
                        if dat[i, j] != meteva.base.IV:
                            dvalue[i, j] = dat[i, j] - top_value

                dvalue[dat == meteva.base.IV] = meteva.base.IV
                picture_ele_dict["error"] = dvalue.tolist()

                if cmap_error is None:
                    cmap_error = "me_bwr"
                cmap_part_e, clevs_e = meteva.base.color_tools.def_cmap_clevs(cmap_error, vmin=-maxd, vmax=maxd)

                meteva.base.tool.myheatmap(ax1, dvalue.T, cmap=cmap_part_e, clevs=clevs_e, annot=annot,fontsize=annot_size)
                #sns.heatmap(dvalue.T, ax=ax1, mask=mask, cmap=cmap_error, vmin=-maxd, vmax=maxd, center=None, robust=False, annot=annot,
                #             fmt=fmt_str, annot_kws={'size':annot_size})

                #ax1.set_xlabel('实况时间',fontsize = sup_fontsize = 0.9)
                ax1.set_ylabel('起报时间',fontsize = sup_fontsize * 0.9)
                ax1.set_xticks(x_plot)
                ax1.set_xticklabels(x_ticks,rotation=360,fontsize=sup_fontsize * 0.8)
                ax1.set_yticks(y_plot)
                ax1.set_yticklabels(y_ticks, rotation=360, fontsize=sup_fontsize * 0.8)

                if isinstance(title,list):
                    title1 = title[kk2]
                    kk2 +=1
                else:
                    if id in meteva.base.station_id_name_dict.keys():
                        title1 = title+"（误差）"+"("+data_name+")"+ "{\'id\':"+str(id)+meteva.base.station_id_name_dict[id] +"}"
                    else:
                        title1 = title + "（误差）" + "(" + data_name + ")" + "{\'id\':" + str(id) + "}"
                ax1.set_title(title1, loc='left', fontweight='bold', fontsize=sup_fontsize)

                ax1.grid(linestyle='--', linewidth=0.5)

                #plt.tick_params(top='on', right='on', which='both')  # 显示上侧和右侧的刻度
                plt.rcParams['xtick.direction'] = 'in'  # 将x轴的刻度线方向设置抄向内
                plt.rcParams['ytick.direction'] = 'in'  # 将y轴的刻度方知向设置向内
                picture_ele_dict["ob_rect"] = []

                for k in range(row + 1):
                    jr = row - k - 1
                    time_fo = times_fo[0] + np.timedelta64(1, 'h') * dh_y * jr
                    dhx0 = (time_fo- times_ob[0]) / np.timedelta64(1, 'h') + min_dtime
                    x1 = (dhx0 - dh_y) / dh_x
                    y1 = k
                    rect = patches.Rectangle((x1-0.5, y1-0.5), dh_y / dh_x, 1, linewidth=2, edgecolor='k', facecolor='none')
                    picture_ele_dict["ob_rect"].append([x1,y1,dh_y/dh_x,1])
                    ax1.add_patch(rect)

                rect = patches.Rectangle((-0.5, -0.5), col, row, linewidth=0.8, edgecolor='k', facecolor='none')
                ax1.set_ylim(row-0.5, -0.5)
                ax1.add_patch(rect)
            else:
                if height is None:
                    height = width * row / col + 1.2
                f, ax2 = plt.subplots(figsize=(width, height), nrows=1, edgecolor='black',dpi = dpi)
                plt.subplots_adjust(left=0.1, bottom=0.15, right=0.98, top=0.90)

            if cmap is None:
                cmap = "rainbow"
            cmap_part ,clev_part= meteva.base.tool.color_tools.def_cmap_clevs(cmap,clev,vmin,vmax)
            # sns.heatmap(dat.T, ax=ax2, mask=mask, cmap=cmap_part, vmin=vmin, vmax=vmax, center=None, robust=False, annot=annot,fmt='.0f'
            # , annot_kws = {'size': annot_size})

            meteva.base.tool.myheatmap(ax2, dat.T, cmap = cmap_part,clevs=clev_part,  annot=annot, fontsize=annot_size)
            ax2.set_xlabel('实况时间',fontsize = sup_fontsize * 0.9)
            ax2.set_ylabel('起报时间',fontsize = sup_fontsize * 0.9)
            ax2.set_xticks(x_plot)
            ax2.set_xticklabels(x_ticks,rotation=360, fontsize=sup_fontsize * 0.8)
            ax2.set_yticks(y_plot)
            ax2.set_yticklabels(y_ticks, rotation=360, fontsize=sup_fontsize * 0.8)
            ax2.grid(linestyle='--', linewidth=0.5)

            ax2.set_ylim(row-0.5,-0.5)
            s1 = s
            if s1 is None:
                s1 = {}
                s1["id"] = id
                s1["member"] =[data_name]
            #title1 = meteva.product.program.get_title_from_dict(meteva.product.time_list_mesh, s1, None, None,None)

            #title = data_name + '实况和不同时效预报对比图'
            if isinstance(title,list):
                title1 = title[kk2]
                kk2 +=1
            else:
                if id in meteva.base.station_id_name_dict.keys():
                    title1 = title + "（要素值）" + "(" + data_name + ")" + "{\'id\':" + str(id) +meteva.base.station_id_name_dict[id] +"}"
                else:
                    title1 = title + "（要素值）" + "(" + data_name + ")" + "{\'id\':" + str(id)  + "}"
            ax2.set_title(title1, loc='left', fontweight='bold', fontsize=sup_fontsize)

            for k in range(row):
                jr = row - k - 1
                dhx0 = (times_fo[0] - times_ob[0]) / np.timedelta64(1, 'h') +min_dtime + dh_y * jr
                x1 = (dhx0-dh_y)/dh_x
                y1 = k
                rect = patches.Rectangle((x1-0.5, y1-0.5), dh_y/dh_x, 1, linewidth=2, edgecolor='k', facecolor='none')
                ax2.add_patch(rect)
            rect = patches.Rectangle((-0.5,-0.5 ), col, row, linewidth=0.8, edgecolor='k', facecolor='none')
            ax2.add_patch(rect)
            #plt.tick_params(top='on', right='on', which='both')  # 显示上侧和右侧的刻度
            plt.rcParams['xtick.direction'] = 'in'  # 将x轴的刻度线方向设置抄向内
            plt.rcParams['ytick.direction'] = 'in'  # 将y轴的刻度方知向设置向内

            save_path1 = None
            if (save_path is None):
                if save_dir is None:
                    show = True
                else:
                    save_path1 = save_dir + "/" + data_name + "_" + str(id) + ".png"
            else:
                save_path1 = save_path[kk1]
            if save_path1 is not None:
                meteva.base.tool.path_tools.creat_path(save_path1)
                figure = plt.gcf()
                plt.tight_layout()
                figure.savefig(save_path1, bbox_inches='tight')
                print("图片已保存至" + save_path1)
            if show:
                plt.show()
            plt.close()
            json_path1 = None
            if json_path is None:
                if json_dir is None:
                    pass
                else:
                    json_path1 = json_dir + "/" + data_name + "_" + str(id) + ".json"
            else:
                json_path1 = json_path[kk1]

            if json_path1 is not None:
                meteva.base.tool.path_tools.creat_path(json_path1)
                file = open(json_path1, "w")
                json.dump(picture_ele_dict, file)
                print("have printed pictrue elements to " + json_path1)
            kk1 += 1
    return



def time_list_mesh_temp(sta_ob_and_fos0,s = None,save_dir = None,save_path = None,plot_error = True,show = False,dpi = 300,annot =0,
                        title = "温度预报准确性和稳定性对比图",
                        sup_fontsize = 10,width = None,height = None,json_dir = None,json_path = None):
    cmap,clev= meteva.base.tool.color_tools.get_cmap_and_clevs_by_element_name("temp")
    time_list_mesh(sta_ob_and_fos0,s,save_dir,save_path,clev,cmap,plot_error,cmap_error= "bwr",show = show,dpi = dpi ,annot = annot,
    title = title,sup_fontsize= sup_fontsize,width=width,height=height,json_dir = json_dir,json_path = json_path)

def time_list_mesh_rain01h(sta_ob_and_fos0,s = None,save_dir = None,save_path = None,plot_error = True,show = False,dpi = 300,annot =0,
                           title = "1小时降水量预报准确性和稳定性对比图",
                           sup_fontsize = 10,width = None,height = None,json_dir = None,json_path = None):
    cmap,clev= meteva.base.tool.color_tools.get_cmap_and_clevs_by_element_name("rain_1h")
    #clev_error, cmap_error = meteva.base.tool.color_tools.get_clev_and_cmap_by_element_name("rain_1h_error")
    time_list_mesh(sta_ob_and_fos0,s,save_dir,save_path,clev,cmap,plot_error,show = show,xtimetype="right",dpi = dpi ,annot = annot,
    title = title,sup_fontsize= sup_fontsize,width=width,height=height,json_dir = json_dir,json_path = json_path)

def time_list_mesh_rain03h(sta_ob_and_fos0,s = None,save_dir = None,save_path = None,plot_error = True,show = False,dpi = 300,annot =0,
                           title = "3小时降水量预报准确性和稳定性对比图",
                           sup_fontsize = 10,width = None,height = None,json_dir = None,json_path = None):
    cmap,clev= meteva.base.tool.color_tools.get_cmap_and_clevs_by_element_name("rain_3h")
    #clev_error, cmap_error = meteva.base.tool.color_tools.get_clev_and_cmap_by_element_name("rain_3h_error")
    time_list_mesh(sta_ob_and_fos0, s, save_dir, save_path, clev, cmap, plot_error, show=show,
                    xtimetype="right",dpi = dpi ,annot = annot,
    title = title,sup_fontsize= sup_fontsize,width=width,height=height,json_dir = json_dir,json_path = json_path)

def time_list_mesh_rh(sta_ob_and_fos0,s = None,save_dir = None,save_path = None,plot_error = True,show = False,dpi = 300,annot =0,
                      title = "相对湿度预报准确性和稳定性对比图",
                      sup_fontsize = 10,width = None,height = None,json_dir = None,json_path = None):
    cmap,clev= meteva.base.tool.color_tools.get_cmap_and_clevs_by_element_name("rh")
    #clev_error, cmap_error = meteva.base.tool.color_tools.get_clev_and_cmap_by_element_name("rh_error")
    time_list_mesh(sta_ob_and_fos0,s,save_dir,save_path, clev, cmap, plot_error,show = show,dpi = dpi ,annot = annot,
    title = title,sup_fontsize= sup_fontsize,width=width,height=height,json_dir = json_dir,json_path = json_path)

def time_list_mesh_vis(sta_ob_and_fos0,s = None,save_dir = None,save_path = None,plot_error = True,show = False,dpi = 300,annot =0,
                       title = "能见度预报准确性和稳定性对比图",
                       sup_fontsize = 10,width = None,height = None,json_dir = None,json_path = None):
    cmap,clev= meteva.base.tool.color_tools.get_cmap_and_clevs_by_element_name("vis")
    #clev_error,cmap_error = meteva.base.tool.color_tools.get_clev_and_cmap_by_element_name("vis_error")
    time_list_mesh(sta_ob_and_fos0,s,save_dir,save_path,clev,cmap,plot_error,show = show,dpi = dpi ,annot = annot,
    title = title,sup_fontsize= sup_fontsize,width=width,height=height,json_dir = json_dir,json_path = json_path)


def time_list_mesh_tcdc(sta_ob_and_fos0,s = None,save_dir = None,save_path = None,plot_error = True,show = False,dpi = 300,annot =0,
                        title = "云量预报准确性和稳定性对比图",
                        sup_fontsize = 10,width = None,height = None,json_dir = None,json_path = None):
    cmap,clev= meteva.base.tool.color_tools.get_cmap_and_clevs_by_element_name("tcdc")
    #clev_error, cmap_error = meteva.base.tool.color_tools.get_clev_and_cmap_by_element_name("tcdc_error")
    time_list_mesh(sta_ob_and_fos0,s,save_dir,save_path,clev,cmap,plot_error = plot_error,show = show,dpi = dpi ,annot = annot,
    title = title,sup_fontsize= sup_fontsize,width=width,height=height,json_dir = json_dir,json_path = json_path)


def time_list_mesh_wind(sta_ob_and_fos0,s = None,save_dir = None,save_path = None,plot_error = True,
                        max_error = None,show = False,dpi = 300,title = "风预报准确性和稳定性对比图",
                        sup_fontsize = 10,width = None,height = None,json_dir = None,json_path = None):

    if max_error is None:
        sta_ob_fos0_noIV = meteva.base.not_IV(sta_ob_and_fos0)
        values = sta_ob_fos0_noIV.values[:,6:].T
        if(values.size ==0):
            print("无有效的观测数据")
            return
        u = values[0::2,:]
        v = values[1::2,:]
        s2 = u * u +v * v
        speed = np.sqrt(s2.astype(np.float32))
        dvalues = speed[1:, :] - speed[0, :]
        maxd = np.max(np.abs(dvalues))
    else:
        maxd = max_error


    sta_ob_and_fos1 = meteva.base.sele_by_dict(sta_ob_and_fos0, s)
    ids = list(set(sta_ob_and_fos1.loc[:, "id"]))
    ids.sort()
    data_names = meteva.base.get_stadata_names(sta_ob_and_fos1)
    ob_names = data_names[0:2]
    fo_names = data_names[2:]
    sta_ob_all1 = meteva.base.sele_by_para(sta_ob_and_fos1,member=ob_names)
    sta_fo_all1 = meteva.base.sele_by_para(sta_ob_and_fos1, member=fo_names)

    times_fo = sta_fo_all1.loc[:, "time"].values
    times_fo = list(set(times_fo))
    if (len(times_fo) == 1):
        print("仅有单个起报时间的预报，程序退出")
        return
    times_fo.sort()
    times_fo = np.array(times_fo)
    # print(times_fo)

    dhs_fo = (times_fo[1:] - times_fo[0:-1])
    if isinstance(dhs_fo[0], np.timedelta64):
        dhs_fo = dhs_fo / np.timedelta64(1, 'h')
    else:
        dhs_fo = dhs_fo / datetime.timedelta(hours=1)
    dhs_fo_not0 = dhs_fo[dhs_fo != 0]
    dh_y = np.min(dhs_fo_not0)
    min_dtime = int(np.min(sta_fo_all1["dtime"]))
    sta_ob_part1 = meteva.base.between_dtime_range(sta_ob_all1, min_dtime, min_dtime + dh_y - 0.1)
    sta_ob_part2 = meteva.base.move_fo_time(sta_ob_part1, dh_y)

    ob_time_s = sta_fo_all1["time"] + sta_fo_all1["dtime"] * np.timedelta64(1, 'h')
    times_ob = list(set(ob_time_s.values))
    times_ob.sort()
    times_ob = np.array(times_ob)

    dhs_ob = (times_ob[1:] - times_ob[0:-1])
    if isinstance(dhs_ob[0], np.timedelta64):
        dhs_ob = dhs_ob / np.timedelta64(1, 'h')
    else:
        dhs_ob = dhs_ob / datetime.timedelta(hours=1)
    dhs_ob_not0 = dhs_ob[dhs_ob != 0]
    dh_x = np.min(dhs_ob_not0)
    # print(dh_x)
    np.sum(dhs_fo_not0)
    row = int(np.sum(dhs_fo_not0) / dh_y) + 2
    col = int(np.sum(dhs_ob_not0) / dh_x) + 1
    # print(row)
    t_ob = []
    for t in times_ob:
        t_ob.append(meteva.base.all_type_time_to_datetime(t))

    # t_fo =[]
    # for t in times_fo:
    #    t_fo.append(meteva.base.all_type_time_to_datetime(t))

    y_plot = np.arange(row) + 0.5
    y_ticks = []
    t_fo0 = meteva.base.all_type_time_to_datetime(times_fo[0])
    step = int(math.ceil(row / 40))

    if step != 1:
        while step * dh_y % 3 != 0:
            step += 1

    y_plot = np.arange(0, row, step)
    for j in range(0, row, step):
        jr = row - j - 1
        time_fo = t_fo0 + datetime.timedelta(hours=1) * dh_y * jr
        hour = time_fo.hour
        day = time_fo.day
        # if ((j * int(dh_y)) % 3 == 0):
        str1 = str(day) + "日" + str(hour) + "时"
        # else:
        #    str1 = str(hour) + "时"
        # print(str1)
        y_ticks.append(str1)

    if width is None:
        width = 8
    x_plot, x_ticks = meteva.product.get_x_ticks(times_ob, width - 2)
    x_plot /= dh_x
    x = np.arange(col)
    y = np.arange(row)
    nfo = int(len(fo_names)/2)
    nids = len(ids)
    if isinstance(title, list):
        if plot_error:
            if 2 * nids * nfo != len(title):
                print("手动设置的title数目和要绘制的图形数目不一致")
                return
        else:
            if nids * nfo != len(title):
                print("手动设置的title数目和要绘制的图形数目不一致")
                return

    if save_path is not None:
        if isinstance(save_path,str):
            save_path = [save_path]
        if nids * nfo != len(save_path):
            print("手动设置的save_path数目和要绘制的图形数目不一致")
            return
    kk1 = 0
    kk2 = 0

    lenght = 40 * (width / col)
    lenght =1.7* lenght**0.5
    for d in range(nfo):
        data_name = fo_names[d*2:d*2+2]
        sta_fo_all2 = meteva.base.in_member_list(sta_fo_all1, data_name)
        meteva.base.set_stadata_names(sta_ob_part2, data_name)
        sta_one_member = meteva.base.combine_join(sta_ob_part2, sta_fo_all2)
        # 以最近的预报作为窗口中间的时刻
        for id in ids:
            picture_ele_dict = {}
            picture_ele_dict["xticklabels"] = meteva.product.program.fun.get_time_str_list(times_ob, row=3)
            picture_ele_dict["subplots"] = {}

            sta_one_id = meteva.base.in_id_list(sta_one_member, id)
            #dat = np.ones((col, row)) * meteva.base.IV
            dat_u = np.ones((row,col)) * meteva.base.IV
            dat_v = np.ones(dat_u.shape)* meteva.base.IV
            for j in range(row):
                jr = row - j - 1
                time_fo = times_fo[0] + np.timedelta64(1, 'h') * dh_y * jr
                #print(time_fo)
                sta_on_row = meteva.base.in_time_list(sta_one_id, time_fo)
                dhx0 = (time_fo - times_ob[0]) / np.timedelta64(1, 'h')
                dhxs = sta_on_row["dtime"].values + dhx0
                index_i = (dhxs / dh_x).astype(np.int16)

                dat_u[j,index_i] = sta_on_row.values[:, -2]
                dat_v[j,index_i] = sta_on_row.values[:, -1]
            dat_speed = np.sqrt(dat_u * dat_u + dat_v*dat_v)
            dat_speed[dat_u == meteva.base.IV] = meteva.base.IV
            mask = np.zeros_like(dat_speed)
            mask[dat_speed == meteva.base.IV] = True

            if plot_error:
                if height is None:
                    height = (width * row / col + 2) * 2
                f, (ax1, ax2) = plt.subplots(figsize=(width, height), nrows=2, edgecolor='black', dpi=dpi)
                plt.subplots_adjust(left=0.1, bottom=0.15, right=0.98, top=0.90, hspace=0.3)

                #lenght = 5*(height / row)
                #print(lenght)
                diff_speed = np.zeros_like(dat_speed)
                diff_u = np.zeros_like(dat_u)
                diff_v = np.zeros_like(dat_v)

                # "风速误差"
                for i in range(col):
                    top_value = meteva.base.IV
                    for j in range(row):
                        if dat_speed[j, i] != meteva.base.IV:
                            top_value = dat_speed[j, i]
                            break
                    for j in range(row):
                        if dat_speed[j, i] != meteva.base.IV:
                            diff_speed[j, i] = dat_speed[j, i] - top_value
                # u 分量误差
                for i in range(col):
                    top_value = meteva.base.IV
                    for j in range(row):
                        if dat_u[j, i] != meteva.base.IV:
                            top_value = dat_u[j, i]
                            break
                    for j in range(row):
                        if dat_u[j, i] != meteva.base.IV:
                            diff_u[j, i] = dat_u[j, i] - top_value
                # v 分量误差
                for i in range(col):
                    top_value = meteva.base.IV
                    for j in range(row):
                        if dat_v[j, i] != meteva.base.IV:
                            top_value = dat_v[j, i]
                            break
                    for j in range(row):
                        if dat_v[j, i] != meteva.base.IV:
                            diff_v[j, i] = dat_v[j, i] - top_value


                cmap_part_e, clevs_e = meteva.base.color_tools.def_cmap_clevs("me_bwr", vmin=-maxd, vmax=maxd)

                meteva.base.tool.myheatmap(ax1, diff_speed, cmap=cmap_part_e, clevs=clevs_e, annot=None)

                #clev, cmap_error = meteva.base.tool.color_tools.get_clev_and_cmap_by_element_name("wind_speed_error")

                #sns.heatmap(diff_speed, ax=ax1, mask=mask, cmap="bwr", vmin=-maxd, vmax=maxd)
                # sns.heatmap(dvalue.T, ax=ax1, mask=mask, cmap=cmap_error, vmin=-maxd, vmax=maxd, center=None, robust=False, annot=True,
                #            fmt=fmt_str)
                #ax1.set_xlabel('实况时间',fontsize =16 )
                ax1.set_ylabel('起报时间',fontsize =sup_fontsize * 0.9)
                ax1.set_xticks(x_plot)
                ax1.set_xticklabels(x_ticks,fontsize =sup_fontsize * 0.8)
                ax1.set_yticks(y_plot)
                ax1.set_yticklabels(y_ticks, rotation=360,fontsize =sup_fontsize * 0.8)
                #title = "实况(id:" + str(id) + ")和不同时效预报(" + data_name[0][2:] + ")偏差图"
                if isinstance(title,list):
                    title1 = title[kk2]
                    kk2 += 1
                else:

                    if id in meteva.base.station_id_name_dict.keys():
                        title1 = title + "(偏差)" + "(" + data_name[0][2:] + ")" + "{\'id\':" + str(id) + meteva.base.station_id_name_dict[id] + "}"
                    else:
                        title1 = title + "(偏差)" + "(" + data_name[0][2:] + ")" + "{\'id\':" + str(id) + "}"

                ax1.set_title(title1, loc='left', fontweight='bold', fontsize=sup_fontsize)
                ax1.grid(linestyle='--', linewidth=0.5)
                xx, yy = np.meshgrid(x , y )
                speed_1d = dat_speed.flatten()
                xx_1d = xx.flatten()[speed_1d != meteva.base.IV]
                yy_1d = yy.flatten()[speed_1d != meteva.base.IV]
                u_1d = diff_u.flatten()[speed_1d != meteva.base.IV]
                v_1d = diff_v.flatten()[speed_1d != meteva.base.IV]
                ax1.barbs(xx_1d, yy_1d, u_1d, v_1d, barb_increments={'half': 2, 'full': 4, 'flag': 20},
                          length=lenght,sizes = dict(emptybarb=0.01, spacing=0.23, height=0.5, width=0.25),
                      linewidth = lenght * lenght * 0.025)

                plt.rcParams['xtick.direction'] = 'in'  # 将x轴的刻度线方向设置抄向内
                plt.rcParams['ytick.direction'] = 'in'  # 将y轴的刻度方知向设置向内
                for k in range(row-1):
                    jr = row - k - 1
                    #dhx0 = (times_fo[jr] - times_ob[0]) / np.timedelta64(1, 'h') + min_dtime
                    #time_fo = times_fo[0] + np.timedelta64(1, 'h') * dh_y * jr
                    #dhx0 = (times_fo[jr] - times_ob[0]) / np.timedelta64(1, 'h') + min_dtime
                    dhx0 = (times_fo[0] - times_ob[0]) / np.timedelta64(1, 'h') + min_dtime + dh_y * jr
                    x1 = (dhx0 - dh_y) / dh_x
                    y1 = k
                    rect = patches.Rectangle((x1-0.5, y1-0.5), dh_y / dh_x, 1, linewidth=2, edgecolor='k', facecolor='none')
                    ax1.add_patch(rect)
                rect = patches.Rectangle((-0.5, -0.5), col, row, linewidth=0.8, edgecolor='k', facecolor='none')
                #ax1.add_patch(rect)
                ax1.set_ylim(row - 0.5, -0.5)
            else:
                if height is None:
                    height = width * row / col + 1.2
                f, ax2 = plt.subplots(figsize=(width, height), nrows=1, edgecolor='black', dpi=dpi)
                plt.subplots_adjust(left=0.1, bottom=0.15, right=0.98, top=0.90)

                #lenght = (height / row)
                #print(lenght)
            vmin = np.min(dat_speed[dat_speed != meteva.base.IV])
            vmax = np.max(dat_speed[dat_speed != meteva.base.IV])
            cmap,clev= meteva.base.tool.color_tools.get_cmap_and_clevs_by_element_name("wind_speed")
            cmap_part,clev_part  = meteva.base.tool.color_tools.get_part_cmap_and_clevs(cmap,clev, vmax, vmin)
            vmax = clev_part[-1]
            vmin = 2 * clev_part[0] - clev_part[1]

            #sns.heatmap(dat_speed, ax=ax2, mask=mask, cmap=cmap_part, vmin=vmin, vmax=vmax)

            cmap_part ,clev_part= meteva.base.tool.color_tools.def_cmap_clevs("wind_speed")
            # sns.heatmap(dat.T, ax=ax2, mask=mask, cmap=cmap_part, vmin=vmin, vmax=vmax, center=None, robust=False, annot=annot,fmt='.0f'
            # , annot_kws = {'size': annot_size})


            meteva.base.tool.myheatmap(ax2, dat_speed, cmap = cmap_part,clevs=clev_part,  annot=None)

            ax2.set_xlabel('实况时间',fontsize =sup_fontsize * 0.9)
            ax2.set_ylabel('起报时间',fontsize =sup_fontsize * 0.8)
            ax2.set_xticks(x_plot)
            ax2.set_xticklabels(x_ticks,fontsize =sup_fontsize * 0.8)
            ax2.set_yticks(y_plot)
            ax2.set_yticklabels(y_ticks, rotation=360,fontsize =sup_fontsize * 0.8)
            #title = "实况(" + str(id) + ")和不同时效预报(" + data_name[0][2:] + ")对比图"
            if isinstance(title,list):
                title1 = title[kk2]
                kk2+= 1
            else:
                if id in meteva.base.station_id_name_dict.keys():
                    title1 = title + "(要素值)" + "(" + data_name[0][2:] + ")" + "{\'id\':" + str(id) + \
                             meteva.base.station_id_name_dict[id] + "}"
                else:
                    title1 = title + "(要素值)" + "(" + data_name[0][2:] + ")" + "{\'id\':" + str(id)  +"}"

            ax2.set_ylim(row - 0.5, -0.5)
            ax2.set_title(title1, loc='left', fontweight='bold', fontsize=sup_fontsize)
            ax2.grid(linestyle='--', linewidth=0.5)
            xx, yy = np.meshgrid(x , y )
            speed_1d = dat_speed.flatten()
            xx_1d = xx.flatten()[speed_1d != meteva.base.IV]
            yy_1d = yy.flatten()[speed_1d != meteva.base.IV]
            u_1d = dat_u.flatten()[speed_1d != meteva.base.IV]
            v_1d = dat_v.flatten()[speed_1d != meteva.base.IV]

            ax2.barbs(xx_1d, yy_1d, u_1d, v_1d, barb_increments={'half': 2, 'full': 4, 'flag': 20},
                      length=lenght,sizes = dict(emptybarb=0.01, spacing=0.23, height=0.5, width=0.25),
                      linewidth = lenght * lenght * 0.025)

            for k in range(row-1):
                jr = row - k - 1
                dhx0 = (times_fo[0] - times_ob[0]) / np.timedelta64(1, 'h') +min_dtime + dh_y * jr
                x1 = (dhx0-dh_y)/dh_x
                y1 = k
                rect = patches.Rectangle((x1-0.5, y1-0.5), dh_y/dh_x, 1, linewidth=2, edgecolor='k', facecolor='none')
                ax2.add_patch(rect)
            #rect = patches.Rectangle((-0.5,-0.5), col, row, linewidth=0.8, edgecolor='k', facecolor='none')
            #ax2.add_patch(rect)

            save_path1 = None
            if (save_path is None):
                if save_dir is None:
                    show = True
                else:
                    save_path1 = save_dir + "/" + data_name[0][2:] + "_" + str(id) + ".png"
            else:
                save_path1 = save_path[kk1]
            if save_path1 is not None:
                meteva.base.tool.path_tools.creat_path(save_path1)
                plt.savefig(save_path1, bbox_inches='tight')
                print("图片已保存至" + save_path1)
            if show:
                plt.show()
            plt.close()
            json_path1 = None
            if json_path is None:
                if json_dir is None:
                    pass
                else:
                    json_path1 = json_dir + "/" + str(id) + ".png"
            else:
                json_path1 = json_path[kk1]

            if json_path1 is not None:
                meteva.base.tool.path_tools.creat_path(json_path1)
                file = open(json_path1, "w")
                json.dump(picture_ele_dict, file)
                print("have printed pictrue elements to " + json_path1)
            kk1 += 1
    return





import numpy as np
from meteva.base.tool.math_tools import ss_iteration
from meteva.base import IV
from meteva.method.yes_or_no.score import pofd_hfmc,pod_hfmc


def roc_auc_score(ob,fo):
    ob = ob[np.argsort(-fo)]
    #fo = fo[np.argsort(-fo)]
    ob_cs = np.cumsum(ob)
    pod = ob_cs/ob_cs[-1]
    pofd = (np.arange(1,ob.size+1) -ob_cs)/(ob.size - ob_cs[-1])
    #plt.plot(pofd_rate,hit_rate)  roc曲线
    delta_pofd = pofd[1:] - pofd[:-1]
    mean_pod = (pod[1:]+pod[:-1])/2
    score = np.sum(mean_pod * delta_pofd)
    return score

def tems(Ob, Fo):
    '''
    计算bs评分的中间结果
    :param Ob:
    :param Fo:
    :return:
    '''
    Ob_shape = Ob.shape
    Fo_shape = Fo.shape
    tems_list = []
    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')
        return
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo = Fo.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo.shape

    for line in range(new_Fo_shape[0]):
        count = Ob.size
        mx = np.mean(Ob)
        sxx = np.mean(np.power(Ob - mx, 2))
        error2 = np.sum(np.power(new_Fo[line,:] - Ob, 2))
        tems_list.append(np.array([count, error2, mx, sxx]))
    tems_array = np.array(tems_list)
    shape = list(Fo_shape[:ind])
    shape.append(4)
    tems_array = tems_array.reshape(shape)
    return tems_array

def tems_merge(tems0, tems1):
    '''
    :param tems0:
    :param tems1:
    :return:
    '''
    tems_array_list = []
    tems0_shape = list(tems0.shape)
    tems1_shape = list(tems1.shape)
    if tems0_shape != tems1_shape:
        print('tems0和tems0维度不匹配')
        return
    tems0 = tems0.reshape((-1, 4))
    tems1 = tems1.reshape((-1, 4))
    new_tmmsss1_shape = tems1.shape
    for line in range(new_tmmsss1_shape[0]):
        tems1_piece = tems1[line, :]
        tems0_piece = tems0[line, :]

        count_total, mx_total, sxx_total = ss_iteration(tems0_piece[0], tems0_piece[2], tems0_piece[3], tems1_piece[0],
                                                        tems1_piece[2], tems1_piece[3])
        error_total = tems0_piece[1] + tems1_piece[1]
        tems_array_list.append(np.array([count_total, error_total, mx_total, sxx_total]))
    tems_array = np.array(tems_array_list)
    tems_array = tems_array.reshape(tems0_shape)
    return tems_array
def bs(Ob, Fo):
    '''
    brier_score 评分
    :param Ob: 输入的概率化实况，多维的numpy，发生了则取值为1，未发生则取值为0
    :param Fo: 预报的概率值，多维的numpy
    :return: 实数形式的评分值
    '''
    '''
    Ob_shape = Ob.shape
    Fo_shape = Fo.shape
    bs_list = []
    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')

        return
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo = Fo.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo.shape

    for line in range(new_Fo_shape[0]):
        #bs0 = brier_score_loss(Ob.flatten(), new_Fo[line, :].flatten())
        bs0 = np.mean((Ob-new_Fo[line, :])**2)
        bs_list.append(bs0)
    if len(bs_list) == 1:
        bs_array = bs_list[0]
    else:
        bs_array = np.array(bs_list)
        shape = list(Fo_shape[:ind])
        bs_array = bs_array.reshape(shape)
        
    '''
    tems_array = tems(Ob,Fo)
    bs_array = bs_tems(tems_array)
    return bs_array

def bs_tems(tems_array):
    '''

    :param tems_array:
    :return:
    '''
    total_count = tems_array[...,0]
    e2_sum = tems_array[...,1]
    brier = e2_sum / total_count
    return brier



def bss(Ob,Fo):
    '''
    :param Ob: 输入的概率化实况，多维的numpy，发生了则取值为1，未发生则取值为0
    :param Fo: 预报的概率值，多维的numpy
    :return: 实数形式的评分值
    '''
    p_climate = np.sum(Ob)/Ob.size
    Fo_climate = np.ones_like(Ob) * p_climate
    bs0 = bs(Ob,Fo)
    bs_climate = bs(Ob,Fo_climate)
    if bs_climate !=0:
        bss0 = 1 - bs0/bs_climate
    else:
        if bs0 ==0:
            bss0 = 1
        else:
            bss0 = IV
    return bss0

def bss_tems(tems_array):
    bs0 = bs_tems(tems_array)
    sxx_total = tems_array[...,3]

    bs_climate = sxx_total
    if bs_climate.size == 1:
        if bs_climate !=0:
            bss0 = 1 - bs0/bs_climate
        else:
            if bs0 ==0:
                bss0 = 1
            else:
                bss0 = IV
    else:
        under = np.zeros_like(bs_climate)
        under[...] = bs_climate[...]
        under[bs_climate == 0] = 1
        bss0 = 1 - bs0 / under
        bss0[bs_climate ==0] = IV
    return bss0

def roc_auc(Ob, Fo):
    '''
    :param Ob: 输入的概率化实况，多维的numpy，发生了则取值为1，未发生则取值为0
    :param Fo: 预报的概率值，多维的numpy
    :return: 实数形式的评分值
    '''
    ob = Ob.flatten()
    Ob_shape = Ob.shape
    Fo_shape = Fo.shape
    roc_auc_list = []
    Ob_shpe_list = list(Ob_shape)
    size = len(Ob_shpe_list)
    ind = -size
    Fo_Ob_index = list(Fo_shape[ind:])
    if Fo_Ob_index != Ob_shpe_list:
        print('预报数据和观测数据维度不匹配')

        return
    Ob_shpe_list.insert(0, -1)
    new_Fo_shape = tuple(Ob_shpe_list)
    new_Fo = Fo.reshape(new_Fo_shape)
    new_Fo_shape = new_Fo.shape

    for line in range(new_Fo_shape[0]):
        roc_auc_list.append(roc_auc_score(ob, new_Fo[line, :].flatten()))
    if len(roc_auc_list) == 1:
        roc_auc_array = roc_auc_list[0]
    else:
        roc_auc_array = np.array(roc_auc_list)
        shape = list(Fo_shape[:ind])
        roc_auc_array = roc_auc_array.reshape(shape)
    return roc_auc_array




def roc_auc_hnh(hnh_array):
    '''

    :param hnh_array:
    :return:
    '''
    ngrade = hnh_array.shape[-2]
    total_grade_num = hnh_array[...,:,0]
    observed_grade_num = hnh_array[...,:,1]
    shape = list(total_grade_num.shape)
    shape.append(4)
    shape = tuple(shape)
    hfmc = np.zeros(shape)
    total_hap = np.sum(observed_grade_num,axis= -1)
    total_num = np.sum(total_grade_num,axis= -1)
    sum_axis = len(observed_grade_num.shape) - 1
    for i in range(ngrade):
        hfmc[...,i, 0] = np.sum(observed_grade_num[...,i:],axis=sum_axis)
        hfmc[...,i, 1] = np.sum(total_grade_num[...,i:],axis=sum_axis) - hfmc[...,i, 0]
        hfmc[...,i, 2] = total_hap - hfmc[...,i, 0]
        hfmc[...,i, 3]= total_num - (hfmc[...,i, 0] + hfmc[...,i, 1]+ hfmc[...,i, 2])

    far = pofd_hfmc(hfmc)
    pod = pod_hfmc(hfmc)

    start1 = np.ones_like(total_num)
    end0 = np.zeros_like(total_num)
    auc = (start1 - far[...,0]) * (start1 + pod[...,0])
    for i in range(1,ngrade):
        auc += (far[...,i-1] - far[...,i]) * (pod[...,i-1] + pod[...,i])
    auc += (far[...,-1] - end0) * (pod[...,-1])
    auc /= 2

    return auc


# -*- coding:UTF-8 -*-
import copy
import numpy as np
import meteva


def sta_data(df,columns = None,
             dtime_units="hour", data_source="", level_type="",
             var_name="", var_cn_name="",
             var_units="", valid_time=0, data_start_columns=6,reset_ids = True):
    '''
    sta_data() 对数据进行格式化成为固定格式
    :param df: dataframe的站点数据
    :param columns: 文件内包含的数据的列名
    :return: 包含‘level', 'time', 'dtime', 'id', 'lon', 'lat',  列的一个dataframe
    '''
    #提取dframe0 列名称
    if columns is None:
        columns = df.columns

    # 将列名变为小写
    columns_1 = []
    for column in columns:
        #column = column.lower()
        columns_1.append(column)
    columns = columns_1

    new_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat']

    # 提取数据列名称,扩展到新df的列名称中
    for column in columns:
        if column not in new_columns:
            new_columns.append(column)
    sta = copy.deepcopy(df)

    sta.columns = columns
    if "id" not in columns:
        sta["id"] = meteva.base.IV
    if reset_ids :reset_id(sta)
    sta.reset_index(inplace=True)

    # 更改列名
    sta = sta.reindex(columns = new_columns)
    #dframe1 = dframe1[new_columns]

    # 排序
    sta.sort_values(by=new_columns[:4],inplace=False)

    if len(sta.columns) == 6:
        sta["data0"] =0
    else:
        #for i in range(6,len(sta.columns)):
        #    sta.iloc[:,i] = (sta.values[:,i]).astype(np.float32)
        pass

    set_stadata_attrs(sta,dtime_units = dtime_units,data_source = data_source,level_type = level_type,
                      var_name = var_name,var_cn_name = var_cn_name,var_units = var_units,valid_time = valid_time,
                      data_start_columns=data_start_columns)

    return sta

def set_stadata_attrs(sta, dtime_units = None,data_source = None,level_type =None,
             var_name = None,var_cn_name = None,
             var_units = None,valid_time = None,data_start_columns = None):

    if sta.attrs is None: sta.attrs = {}
    if dtime_units is not None:sta.attrs["dtime_units"] = dtime_units
    if data_source is not None:sta.attrs["data_source"] = data_source
    if level_type is not None: sta.attrs["data_type"] = level_type
    if var_name is not None: sta.attrs["var_name"] = var_name
    if var_cn_name is not None:sta.attrs["var_cn_name"] = var_cn_name
    if var_units is not None:sta.attrs["var_units"] = var_units
    if valid_time is not None:sta.attrs["valid_time"] = valid_time
    if data_start_columns is not None:sta.attrs["data_start_columns"] = data_start_columns


def get_undim_data_names(sta):
    '''

    :param sta:
    :return:
    '''
    coor_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat']
    columns = sta.columns
    data_columns = []
    for column in columns:
        if column not in coor_columns:
            if column.find("dim_type")!=0:
                data_columns.append(column)
    return data_columns

def get_stadata_names(sta):
    '''
    get_data_names() 获取站点数据的要素名
    :param sta: 站点数据
    :return: 要素名列表
    '''
    #coor_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat']
    #columns = sta.columns
    #data_columns = []
    #for column in columns:
    #    if column not in coor_columns:
    #        data_columns.append(column)
    columns = sta.columns.values
    if "data_start_columns" not in sta.attrs.keys():
        sta.attrs["data_start_columns"] = 6
    data_start_column = sta.attrs["data_start_columns"]
    data_columns = columns[data_start_column:].tolist()
    return data_columns


def get_expanded_coord_names(sta):
    columns = sta.columns.values
    data_start_column = sta.attrs["data_start_columns"]
    ecc = columns[0:data_start_column].tolist()
    return ecc

def get_coord_names():
    '''

    :return: 站点数据基本信息列名['level', 'time', 'dtime', 'id', 'lon', 'lat']列表
    '''
    return ['level', 'time', 'dtime', 'id', 'lon', 'lat']

def set_stadata_names(sta,data_name_list):
    '''
    更改 要素名，和添加缺省列
    :param sta: 站点数据
    :param data_name: 站点数据 要素名
    :return: 更改要素名名后的站点数据
    '''
    if not isinstance(data_name_list,list):
        data_name_list = [data_name_list]
    #if isinstance(data_name_list,list):
    coor_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat']
    for data_name in data_name_list:
        coor_columns.append(data_name)
    sta.columns = coor_columns
    #else:
    #    print("输出名称设置不成功，数据名称列表参数需为list形式")
    return

def set_stadata_coords(sta,level = None,time = None,dtime = None,id = None,lat = None,lon = None):
    '''
    set_time_dtime_level_name 设置time_dtime_level 的值  并且设置要素名
    :param sta: 站点数据
    :param time: 起报时
    :param dtime: 时效
    :param level: 层次
    :param data_name: 要素名
    :return:  站点数据
    '''
    if time is not None:
        time1 = meteva.base.all_type_time_to_time64(time)
        sta.loc[:,'time'] = time1
        if len(sta.index) ==1:
            sta["time"] = sta["time"].astype(np.datetime64)
    if dtime is not None:
        sta.loc[:,'dtime'] = dtime
        dtime_type = str(sta["dtime"].dtype)
        if dtime_type.find("int") < 0:
            sta[["dtime"]] = sta[['dtime']].astype(int)
    if level is not None:
        sta.loc[:,'level'] = level
    if id is not None:
        sta.loc[:,"id"] = id
    if lat is not None:
        sta.loc[:,"lat"] = lat
    if lon is not None:
        sta.loc[:,"lon"] = lon



def reset_id(sta):
    '''
    输入的sta的站号中可能有些站号包含a-z,A-Z的字母，对此将这些字母转换为对应的ASCII数字，再将整个字符串格式的站号转换为数值形式
    返回sta站号为整型
    '''
    #print(sta)

    id_type = str(sta["id"].dtype)
    if id_type.find("int32") <0:
        values = sta['id'].values
        if id_type.find("int") >=0 or id_type.find("float") >=0:
            int_id = values.astype(np.int32)
            sta['id'] = int_id
        else:
            int_id = np.zeros(len(values))
            for i in range(len(values)):
                if isinstance(values[i],str):
                    strs = values[i]
                    if strs =="":
                        strs_int=999999
                    else:
                        strs_int = ""
                        for s in strs:
                            if s.isdigit():
                                strs_int += s
                            else:
                                strs_int += str(ord(s))
                    int_id[i] = int(strs_int)
                else:
                    int_id[i] = values[i]

            int_id = int_id.astype(np.int32)
            sta['id'] = int_id
    return





import numpy as np
from . import mij
import math


def imomenter(x, loc=None):
    x = np.array(x)
    out = {}
    if loc is None:
        xdim = x.shape
        range0 = np.tile(np.arange(xdim[0]), xdim[1])
        range1 = (np.arange(xdim[1])).repeat(xdim[0])
        loc = np.stack((range0, range1), axis=-1)
    m00 = mij.mij(x, s=np.array(loc))
    m10 = mij.mij(x, s=np.array(loc), i=1)
    m01 = mij.mij(x, s=np.array(loc), j=1)
    m11 = mij.mij(x, s=np.array(loc), i=1, j=1)
    m20 = mij.mij(x, s=np.array(loc), i=2)
    m02 = mij.mij(x, s=np.array(loc), j=2)
    xbar = m10 / m00
    ybar = m01 / m00
    cen = {'x': xbar, 'y': ybar}
    mu11 = m11 / m00 - xbar * ybar
    mu20 = m20 / m00 - math.pow(xbar, 2)
    mu02 = m02 / m00 - math.pow(ybar, 2)
    theta = 0.5 * math.atan2(2 * mu11, mu20 - mu02)
    out["area"] = m00
    out["centroid"] = cen
    out["orientation_angle"] = theta
    raw = {"m00": m00, "m10": m10, "m01": m01, "m11": m11, "m20": m20, "m02": m02}
    out["raw.moments"] = raw
    out["cov"] = np.array([[mu20, mu11], [mu11, mu02]])
    out["class"] = "imomented"
    return out



import os
import meteva
import numpy as np
def get_hour_dhour_list(gds_file_list):
    hour_list = []
    dh_list = []
    for path in gds_file_list:
        dati = get_dati_of_path(path)
        hour_list.append(dati.hour)
        file, ftype = os.path.splitext(path)
        dh_list.append(int(ftype[1:]))
    hour_list1 = list(set(hour_list))
    hour_list1.sort()
    dh_list1 = list(set(dh_list))
    dh_list1.sort()
    return hour_list1,dh_list1

def get_dati_of_path(path):
    dir,filename = os.path.split(path)
    filename0 = os.path.splitext(filename)[0]
    a = int(filename0[0:2])
    b = int(filename0[2:4])
    if a ==20:
        if b >12:
            pass
        else:
            filename0 = "20" + filename0
    elif a == 19:
        if b >12:
            pass
        else:
            filename0 = "20" + filename0
    else:
        filename0 = "20" + filename0

    dati = meteva.base.time_tools.str_to_time(filename0)
    return dati
def get_dati_str_of_path(path):
    dir,filename = os.path.split(path)
    filename0 = os.path.splitext(filename)[0]
    a = int(filename0[0:2])
    b = int(filename0[2:4])
    if a ==20:
        if b >12:
            pass
        else:
            filename0 = "20" + filename0
    elif a == 19:
        if b >12:
            pass
        else:
            filename0 = "20" + filename0
    else:
        filename0 = "20" + filename0
    dati_str = filename0[0:8]

    return dati_str


def get_ob_time_range_str(sta_all):

    dtimes = sta_all["dtime"] * np.timedelta64(1, 'h')
    obtimes = sta_all['time'] + dtimes
    time_s = meteva.base.all_type_time_to_datetime(np.min(obtimes))
    time_e = meteva.base.all_type_time_to_datetime(np.max(obtimes))
    strs_list = meteva.product.get_time_str_list([time_s, time_e])
    str1 = strs_list[0] + "—" + strs_list[1]
    return str1

def get_time_range_str(sta_all):
    times = sta_all['time']
    time_s = meteva.base.all_type_time_to_datetime(np.min(times))
    time_e = meteva.base.all_type_time_to_datetime(np.max(times))
    strs_list = meteva.product.get_time_str_list([time_s, time_e])
    str1 = strs_list[0] + "—" + strs_list[1]
    return str1



#python setup.py sdist bdist_wheel
#twine upload dist/*
__author__ = "The R & D Center for Weather Forecasting Technology in NMC, CMA"
__version__ = '1.7.4'

from . import base
from . import method
from . import perspact
from . import product




import json
import meteva

def load_feature_summary(filename):
    file = open(filename, "r")
    strs = file.read()
    feature = json.loads(strs)
    file.close()
    feature1 = {}
    feature1["match_count"] = feature["match_count"]
    feature1["time"] = feature["time"]
    feature1["dtime"] = feature["dtime"]
    feature1["member"] = feature["member"]
    feature1["feature_table"] = feature["feature_table"]
    feature1["interester"] = feature["interester"]


    label_list_matched = feature["label_list_matched"]
    feature1["label_list_matched"] = label_list_matched
    for id in label_list_matched:
        feature1[id] = feature[str(id)]

    feature1["unmatched"] = feature["unmatched"]
    miss_labels = feature["unmatched"]['ob']
    for id in miss_labels:
        feature1[id] = feature[str(id)]


    false_alarm_labels = feature["unmatched"]["fo"]
    for id in false_alarm_labels:
        feature1[id] = feature[str(id)]

    return feature1

def load_feature_summary_list(dir):
    file_list = meteva.base.tool.path_tools.get_path_list_in_dir(dir)
    features_list = []
    for file in file_list:
        feature = load_feature_summary(file)
        features_list.append(feature)
    return features_list


from multiprocessing import cpu_count,Process, freeze_support
import time

def multi_operation(process_id,method,**kwargs):
    print("start running the " + str(process_id) + "th sub process")
    paras = method.__code__.co_varnames

    #找出那些需要循环技术的参数
    list_para_name_dict = {}
    for para_name in kwargs.keys():
        for para in paras:
            if para_name.lower() == para.lower():
                if para_name != para:
                    #list_para_name_dict的key记录kwargs中输入参数名称
                    # list_para_name_dict的value记录method需要的参数名称
                    list_para_name_dict[para_name] = para

    op_count = 0
    for key in list_para_name_dict.keys():
        para_value = kwargs[key]
        op_count = len(para_value)


    num = op_count
    start = time.time()
    kwargs_one = {}

    for key in kwargs.keys():
        if key in list_para_name_dict.keys():
            key_new = list_para_name_dict[key]  # 获取method的实际参数名称
            kwargs_one[key_new] = None
        else:
            kwargs_one[key] = kwargs[key]  # 其它参数照常保留

    for i in range(num):
        for key in list_para_name_dict.keys():
            key_new = list_para_name_dict[key]  # 获取method的实际参数名称
            kwargs_one[key_new] = kwargs[key][i]
        method(**kwargs_one)
        #print("***********")
        if process_id == 0:
            end = time.time()
            left_minutes = int((end - start) * (num - i - 1) / ((i + 1) * 60)) + 1
            print("剩余" + str(left_minutes) + "分钟")



def multi_run(process_count,method,** kwargs):
    # 判断那些需要并行的参数
    #
    freeze_support()
    paras = method.__code__.co_varnames
    list_para_name_dict = {}
    op_count_max = 0
    for para_name in kwargs.keys():
        for para in paras:
            if para_name.lower() == para.lower():
                if para_name != para:
                    #list_para_name_dict的key记录了原始的输入参数名称
                    # list_para_name_dict的value记录了更改后的输入参数名称
                    # 如果输入参数已经和method参数大小写不一致，key 和value就相同
                    list_para_name_dict[para_name] = para_name
                    op_count_max = len(kwargs[para_name])

    if len(list_para_name_dict.keys()) ==0:
        #从所有的参数中，找出参数值为列表的参数，并挑选出列表长度最大值

        for para_name in kwargs.keys():
            para_value = kwargs[para_name]
            if isinstance(para_value,list):
                op_count = len(para_value)
                if op_count_max < op_count:
                    op_count_max = op_count
        if op_count_max ==0:
            print("未指定需要并行的参数列表")

        #认为列表长度最大值对应的参数为需要并行的参数名称
        for para_name in kwargs.keys():
            para_value = kwargs[para_name]
            if isinstance(para_value,list):
                op_count = len(para_value)
                if op_count_max == op_count:
                    #list_para_name_dict的key记录了原始的输入参数名称
                    # list_para_name_dict的value记录了更改后的输入参数名称
                    # 如果输入参数已经和method参数大小一致，则对参数名称的大小写进行转换作为传入下一级函数的输入参数名称
                    list_para_name_dict[para_name] = para_name.swapcase()

    #确定可用或需要的线程数
    if process_count<1:
        process_count = 1
    if process_count>cpu_count() -2:
        process_count = cpu_count() -2
    if process_count> op_count_max:
        process_count = op_count_max

    #将执行参数拆解成threed_count份
    kwargs_list =[]
    for k in range(process_count):
        kwargs_k = {"process_id":k,
                    "method":method}
        for key in kwargs.keys():
            if key in list_para_name_dict.keys():
                key_new = list_para_name_dict[key]  #采用大小写和method中不一致的参数名称
                kwargs_k[key_new] = kwargs[key][k::process_count]  #提取部分待运算的参数列表
            else:
                kwargs_k[key] = kwargs[key]  #其它参数照常保留
        kwargs_list.append(kwargs_k)

    PP = []  # 记录进程
    for k in range(process_count):
        tmpp = Process(target=multi_operation, kwargs=kwargs_list[k])
        PP.append(tmpp)

    print('Waiting for all subprocesses done...')
    for pc in PP:
        pc.start()
    for pp in PP:
        pp.join()
    print('All subprocesses done.')


# -*- coding: utf-8 -*-
"""
Created on Tue Oct 20 16:26:07 2020

@author: 1
"""
import numpy as np
import meteva

#提取连通域分析以后被标记的数据
def pick_labels(data):
    ret = {}
    for item, values in data.items():
        if isinstance(item,int):
            #print(item, values)
            ret0 = {item : values}
            ret.update(ret0)
    return ret

#连通域分析后被标记的目标
def relabeled(x):
    label = np.array(())
    ret = None
    for i in range(len(x)):
        label = x['labels_'+str(i+1)]
        label[label == 1] = i+1
        if ret is None:
            ret = label
        else: 
            ret =ret + label
    return ret

#利用xarray将数据转换为grid格式
def get_grd(dataset_ob, data):
    #dataset_ob参数是为了获取原始场的起始经纬度坐标和间隔
    #整理无坐标信息的格点场的坐标：原始的观测场和预报场的经纬度范围应该是对应的
    glon = [dataset_ob["lon"].values[0], dataset_ob["lon"].values[-1],
            dataset_ob["lon"].values[1]-dataset_ob["lon"].values[0]]    #经度的起点、终点、间隔
    glat = [dataset_ob["lat"].values[0], dataset_ob["lat"].values[-1],
             dataset_ob["lat"].values[1]-dataset_ob["lat"].values[0]]    #纬度的起点、终点、间隔

    grid1 = meteva.base.grid(glon, glat)
    data_tr = meteva.base.grid_data(grid1, data)
    return data_tr 

#调整数组存放结构
def funtion(fo_id_list,ob_id_list):

    n_row = len(ob_id_list)
    ob_index_dict = {}
    fo_index_dict = {}
    index_list_dict = {}
    index = 0

    for i in range(n_row):
        if ob_id_list[i] in ob_index_dict.keys() or fo_id_list[i] in fo_index_dict.keys():
            if ob_id_list[i] in ob_index_dict.keys():
                index1 = ob_index_dict[ob_id_list[i]]
                fo_index_dict[fo_id_list[i]] = index1
            elif fo_id_list[i] in fo_index_dict.keys():
                index1 = fo_index_dict[fo_id_list[i]]
                ob_index_dict[ob_id_list[i]] = index1
        else:
            ob_index_dict[ob_id_list[i]] = index
            fo_index_dict[fo_id_list[i]] = index
            index1 = index
            index_list_dict[index1] = []
            index = index + 1
        index_list_dict[index1].append([fo_id_list[i],ob_id_list[i]])

    list_list = list(index_list_dict.values())

    #将ob,或fo id有重复的列表进行合并
    while True:
        len0 = len(list_list)
        changed = False
        for i in range(len0-1):
            dat_i = np.array(list_list[i])
            ob_seti = set(dat_i[:,0])
            fo_seti = set(dat_i[:,1])
            for j in range(i+1,len0):
                dat_j = np.array(list_list[j])
                ob_setj = set(dat_j[:,0])
                fo_setj = set(dat_j[:,1])
                if len(ob_seti&ob_setj) >0 or  len(fo_seti&fo_setj) >0:
                    list_list[i].extend(list_list[j])
                    list_list.pop(j)
                    changed = True
                    break
            if changed:
                break
        if not changed:
            break
    return list_list

#if __name__ == '__main__':
    #fo_id_list = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 3, 4, 5, 5, 5, 8, 9, 10]
    #ob_id_list = [1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 12, 13, 14, 15, 16, 17, 19, 9, 7, 10, 14, 14, 15, 22, 19, 20, 23]

    #list_list = funtion(fo_id_list,ob_id_list)


import math
import datetime
import re
from copy import deepcopy
import numpy as np

class grid:
    '''
        定义一个格点的类grid，来存储网格的范围包括（起始经纬度、格距、起止时间，时间间隔，起止时效，时效间隔，层次列表，数据成员）
        约定坐标顺序为: member, time,ddtime, level, lat,lon
    '''
    def __init__(self,glon, glat, gtime=None, dtime_list=None,level_list=None,member_list = None):

        #提取成员维度信息
        if(member_list is None):
            self.members =['data0']
        else:
            self.members = member_list
        ############################################################################
        #提取层次维度信息
        if(level_list is None):
            self.levels =[0]
        else:
            self.levels = level_list
        ############################################################################
        #提取时间维度信息
        self.stime = np.datetime64('2099-01-01T00:00:00.000000')
        self.etime = np.datetime64('2099-01-01T00:00:00.000000')
        self.dtime_int = 1
        self.dtime_type = "h"
        self.dtimedelta = np.timedelta64(1,'h')
        if(gtime == None):gtime = []
        if len(gtime) == 1:
            if type(gtime[0]) == str:
                num = ''.join([x for x in gtime[0] if x.isdigit()])
                # 用户输入2019041910十位字符，后面补全加0000，为14位统一处理
                if len(num) == 4:
                    num += "0101000000"
                elif len(num) == 6:
                    num +="01000000"
                elif len(num) == 8:
                    num +="000000"
                elif len(num) == 10:
                    num +="0000"
                elif len(num) == 12:
                    num +="00"
                elif len(num) == 14:
                    pass
                else:
                    print("输入日期有误，请检查！")
                # 统一将日期变为datetime类型
                self.stime = datetime.datetime.strptime(num, '%Y%m%d%H%M%S')
                self.etime = datetime.datetime.strptime(num, '%Y%m%d%H%M%S')
                self.stime = np.datetime64(self.stime)
                self.etime = np.datetime64(self.etime)
            else:
                self.stime = gtime[0]
                self.etime = gtime[0]
            self.dtime_int = 1
            self.dtime_type = "h"
            self.dtimedelta = np.timedelta64(0,'h')
        elif len(gtime) ==3:
            num1 =[]
            if type(gtime[0]) == str:
                for i in range (0,2):
                    num = ''.join([x for x in gtime[i] if x.isdigit()])
                    #用户输入2019041910十位字符，后面补全加0000，为14位统一处理
                    if len(num) == 4:
                        num1.append(num + "0101000000")
                    elif len(num) == 6:
                        num1.append(num + "01000000")
                    elif len(num) == 8:
                        num1.append(num + "000000")
                    elif len(num) == 10:
                        num1.append(num + "0000")
                    elif len(num) == 12:
                        num1.append(num + "00")
                    elif len(num) == 14:
                        num1.append(num)
                    else:
                        print("输入日期有误，请检查！")
                    #统一将日期变为datetime类型
                #print(num1)
                self.stime = datetime.datetime.strptime(num1[0], '%Y%m%d%H%M%S')
                self.etime = datetime.datetime.strptime(num1[1], '%Y%m%d%H%M%S')
                self.stime = np.datetime64(self.stime)
                self.etime = np.datetime64(self.etime)
            elif isinstance(gtime[0],np.datetime64):
                stime = gtime[0].astype(datetime.datetime)
                etime = gtime[1].astype(datetime.datetime)
                if isinstance(stime, int):
                    stime = datetime.datetime.utcfromtimestamp(stime / 1000000000)
                    etime = datetime.datetime.utcfromtimestamp(etime / 1000000000)
                self.stime = stime
                self.etime = etime
            else:
                self.stime = gtime[0]
                self.etime = gtime[1]

            if type(gtime[2]) == str:
                self.dtime_int = re.findall(r"\d+", gtime[2])[0]
                dtime_type = re.findall(r"\D+", gtime[2])[0]
                if dtime_type == 'h':
                    self.dtime_type ="h"
                    self.dtimedelta = np.timedelta64(self.dtime_int,'h')
                elif dtime_type == 'd':
                    self.dtime_type ="D"
                    self.dtimedelta = np.timedelta64(self.dtime_int, 'D')
                elif dtime_type == 'm':
                    self.dtime_type ="m"
                    self.dtimedelta = np.timedelta64(self.dtime_int, 'm')
            elif isinstance(gtime[2],np.timedelta64):
                seconds = int(gtime[2] / np.timedelta64(1, 's'))
                if seconds % 3600 == 0:
                    self.dtime_type = "h"
                    self.dtime_int = int(seconds / 3600)
                else:
                    self.dtime_type = "m"
                    self.dtime_int = int(seconds / 60)
            else:
                self.dtimedelta = gtime[2]
                seconds = gtime[2].total_seconds()
                if seconds % 3600 == 0:
                    self.dtime_type = "h"
                    self.dtime_int = int(seconds/3600)
                else:
                    self.dtime_type = "m"
                    self.dtime_int = int(seconds / 60)
        self.gtime = [self.stime,self.etime,str(self.dtime_int) + self.dtime_type]
        self.stime_str = str(self.stime).replace("-","").replace(" ","").replace(":","").replace("T","")[0:14]
        self.etime_str = str(self.etime).replace("-", "").replace(" ", "").replace(":", "").replace("T", "")[0:14]
        self.dtime_str = str(self.dtime_int) + self.dtime_type

        ############################################################################
        #提取预报时效维度信息
        if dtime_list is None:
            self.dtimes = [0]
        else:
            self.dtimes = dtime_list
        ############################################################################
        #提取经度信息

        self.slon = get_true_value(glon[0])
        self.elon = get_true_value(glon[1])
        self.dlon = get_true_value(glon[2])
        nlon = 1 + (self.elon - self.slon) / self.dlon
        error = abs(round(nlon) - nlon)/nlon
        if (error > 0.01):
            self.nlon = int(math.ceil(nlon))
        else:
            self.nlon = int(round(nlon))

        self.elon = get_true_value(self.slon + (nlon - 1) * self.dlon)
        self.glon = [self.slon,self.elon,self.dlon]

        ############################################################################
        #提取纬度信息
        self.slat = get_true_value(glat[0])
        self.elat = get_true_value(glat[1])
        self.dlat = get_true_value(glat[2])
        nlat = 1 + (self.elat - self.slat) / self.dlat
        error = abs(round(nlat) - nlat)/nlat
        if (error > 0.01):
            self.nlat = int(math.ceil(nlat))
        else:
            self.nlat = int(round(nlat))
        self.elat = get_true_value(self.slat + (nlat - 1) * self.dlat)
        self.glat = [self.slat,self.elat,self.dlat]



    #对原有的格点数据进行一次深拷贝，不改变原有的值和结构。
    def copy(self):
        return deepcopy(self)

    #reset的作用是把网格的坐标间隔统一为正数。
    def reset(self):
        if (self.dlon > 0 and self.dlat > 0):
            pass
        if (self.dlat < 0):
            tran = self.slat
            self.slat = self.elat
            self.elat = tran
            self.dlat = abs(self.dlat)
        if (self.dlon < 0):
            tran = self.slon
            self.slon = self.elon
            self.elon = tran
            self.dlon = abs(self.dlon)
        return


    def __str__(self):
        '''
        重置系统自动的函数，在print(grid) 的时候可以很整齐的看到所有信息
        :return:  string
        '''
        grid_str = ""
        grid_str += "members:" + str(self.members) +"\n"
        grid_str += "levels:" + str(self.levels) + "\n"
        grid_str += "gtime:" + str([self.stime_str,self.etime_str,self.dtime_str]) + "\n"
        grid_str += "dtimes:" + str(self.dtimes)  +"\n"
        grid_str += "glon:" + str(self.glon) + "\n"
        grid_str += "glat:" + str(self.glat) + "\n"
        return grid_str

def get_true_value(value):
    dlon2 = round(value, 2)
    dlon3 = round(value, 3)
    dlon4 = round(value, 4)
    dlon5 = round(value, 5)
    if dlon2 == dlon3 and dlon3 == dlon4:
        return dlon3
    elif dlon3 == dlon4 and dlon5 == dlon4:
        return dlon3
    else:
        return value


def get_grid_of_data(grid_data0):
    '''
     获取grid的数据values值
    :param grid_data0:初始化之后的网格数据
    :return:返回grid数据。
    '''
    member_list = grid_data0['member'].values
    level_list = grid_data0['level'].values
    times = grid_data0['time'].values
    #print(times)
    if(len(times)>1):
        gtime = [times[0],times[-1],times[1]-times[0]]
    elif len(times) == 1:
        gtime = times
    else:
        gtime = None

    gdt = grid_data0['dtime'].values.tolist()
    attrs_name = list(grid_data0.attrs)


    lons = grid_data0['lon'].values

    #dlon5 = round(lons[1] - lons[0], 5)
    #dlon6 = round(lons[1] - lons[0], 6)
    #dlon7 = round(lons[1] - lons[0], 7)
    #if dlon5 == dlon6 and dlon6 == dlon7:
    #    dlon = dlon5
    #else:
    #    dlon = lons[1]-lons[0]
    dlon = get_true_value(lons[1] - lons[0])

    #glon = [lons[0],round(lons[-1],5),round(lons[1]-lons[0],5)]
    glon = [get_true_value(lons[0]), get_true_value(lons[-1]), dlon]
    lats = grid_data0['lat'].values
    #dlat5 = round(lats[1] - lats[0], 5)
    #dlat6 = round(lats[1] - lats[0], 6)
    #dlat7 = round(lats[1] - lats[0], 7)
    #if dlat5 == dlat6 and dlat6 == dlat7:
    #    dlat = dlat5
    #else:
    #    dlat = lats[1]-lats[0]
    dlat = get_true_value(lats[1] - lats[0])

    #glat = [lats[0],round(lats[-1],5),round(lats[1]-lats[0],5)]
    glat = [get_true_value(lats[0]), get_true_value(lats[-1]), dlat]
    grid01 = grid(glon, glat, gtime, gdt, level_list, member_list)
    return grid01


def reset_grid(grid0):
    if grid0.dlat <0:
        grid0.dlat = - grid0.dlat
        tran = grid0.slat
        grid0.slat = grid0.elat
        grid0.elat = tran
    if grid0.dlon <0:
        grid0.dlon = - grid0.dlon
        tran = grid0.slon
        grid0.slon = grid0.elon
        grid0.elon = tran
    return



import meteva


def sample_tdt(sta_all,return_result = False,**kwargs):
    if "title" not in kwargs.keys():
        kwargs["title"] =  ["样本数随时间和时效的分布"]
    if "x_y" not in kwargs.keys():
        kwargs["x_y"] ="time_dtime"
    kwargs["width"] = 10
    sta_all_0 = sta_all.copy()
    sta_all_0.iloc[:,6:] = 0
    result = meteva.product.score_tdt(sta_all_0,meteva.method.sample_count,annot = -1,add_min_xticks = True
                                      ,**kwargs)
    if return_result:
        return result

def sample_id(sta_all,return_result = False,**kwargs):
    if "title" not in kwargs.keys():
        kwargs["title"] =  ["所有时间和时效样本数的空间分布"]
    sta_all_0 = sta_all.copy()
    sta_all_0.iloc[:,6:] = 0
    result = meteva.product.score_id(sta_all_0,meteva.method.sample_count,plot = "scatter",
                                    **kwargs)
    if return_result:
        return result



import pandas as pd
import numpy as np
import meteva
def write_array_to_excel(array,save_path,name_list_dict,columns = None,index = None):
    shape = array.shape
    keys = list(name_list_dict.keys())
    if len(shape) ==1:
        rows = [[keys[0]],name_list_dict[keys[0]]]
        table_data = pd.DataFrame(array,index=pd.MultiIndex.from_product(rows))
        table_data.to_excel(save_path, sheet_name='sheet1')
    elif len(shape)==2:
        dat = None
        if columns is None:
            if index is None:
                columns = keys[0]
                index = keys[1]
            else:
                if index != keys[0]:
                    columns = keys[0]
                else:
                    columns = keys[1]
                    #dat = array.T
        else:
            if columns ==keys[0]:
                index = keys[1]
            else:
                index = keys[0]
        if dat is None:
            dat = array
        if columns not in keys:
            print("legend 参数的取值必须是name_list_dict的key")
        if index not in keys:
            print("axis 参数的取值必须是name_list_dict的key")

        if columns == keys[0]:
            dat = dat.T
        cols = [[columns],name_list_dict[columns]]
        rows = [[index],name_list_dict[index]]
        table_data = pd.DataFrame(dat,columns=pd.MultiIndex.from_product(cols),
                                  index=pd.MultiIndex.from_product(rows))
        table_data.to_excel(save_path, sheet_name='sheet1')
    elif len(shape)==3:
        if name_list_dict is None:
            name_list_dict["z"] = np.arange(shape[0])
            name_list_dict["y"] = np.arange(shape[1])
            name_list_dict["x"] = np.arange(shape[2])
            columns = "y"
            index = "x"
            sheet = "z"
        keys = list(name_list_dict.keys())
        if columns is None:
            if index is None:
                columns = keys[1]
                index = keys[2]
                sheet = keys[0]
            else:
                if index == keys[2]:
                    columns = keys[1]
                    sheet = keys[0]
                elif index == keys[1]:
                    columns = keys[2]
                    sheet = keys[0]
                else:
                    columns = keys[2]
                    sheet = keys[1]
        else:
            if index is None:
                if columns == keys[0]:
                    index = keys[2]
                    sheet = keys[1]
                elif columns == keys[1]:
                    index = keys[2]
                    sheet = keys[0]
                else:
                    index = keys[1]
                    sheet = keys[0]
            else:
                indexlist = [0,1,2]
                indexlist.remove(keys.index(columns))
                indexlist.remove(keys.index(index))
                sheet = keys[indexlist[0]]
        if columns not in keys:
            print("legend 参数的取值必须是name_list_dict的key")
        if index not in keys:
            print("axis 参数的取值必须是name_list_dict的key")
        newshape = (keys.index(sheet),keys.index(index),keys.index(columns))
        data = array.transpose(newshape)
        table_data_list = []
        cols= [[columns],name_list_dict[columns]]
        rows = [[index],name_list_dict[index]]
        sheet_list = name_list_dict[sheet]
        if isinstance(sheet_list[0],pd._libs.tslibs.timestamps.Timestamp):
            print(type(sheet_list[0]))
            sheet_list1 = []
            for name in sheet_list:
                name1 = meteva.base.all_type_time_to_str(name)
                sheet_list1.append(name1)
            sheet_list = sheet_list1
        for s in range(len(sheet_list)):
            # table_data = pd.DataFrame(array[:,:,s])
            table_data = pd.DataFrame(data[s,:, :],
                                      columns=pd.MultiIndex.from_product(cols),
                                      index=pd.MultiIndex.from_product(rows)
                                      )
            table_data_list.append(table_data)
        with pd.ExcelWriter(save_path) as writer:
            for i in range(len(sheet_list)):
                str1 = str(sheet_list[i])
                str1 = str1.replace("[","【")
                #print(str1)
                table_data_list[i].to_excel(writer, sheet_name=sheet + '_' + str1)
    else:
        print("array不能超过3维")
        return


    print("数据已以excel表格形式保存至" + save_path)





import numpy as np
import math
import meteva
import os
import traceback
import warnings
warnings.filterwarnings("ignore")

def write_griddata_to_micaps4(da,save_path = "a.txt",creat_dir = False,effectiveNum = 3,show = False,title = None):
    """
    输出micaps4格式文件
    :param da:xarray多维数据信息
    :param path:存储路径
    :param effectiveNum 有效数字 默认：3
    :return 最终按照需要保存的路径，将da数据保存为m4格式
    """
    try:
        dir = os.path.split(os.path.abspath(save_path))[0]
        if not os.path.isdir(dir):
            if not creat_dir:
                print("文件夹："+dir+"不存在")
                return False
            else:
                meteva.base.tool.path_tools.creat_path(save_path)

        grid = meteva.base.basicdata.get_grid_of_data(da)
        nlon = grid.nlon
        nlat = grid.nlat
        slon = grid.slon
        slat = grid.slat
        elon = grid.elon
        elat = grid.elat
        dlon = grid.dlon
        dlat = grid.dlat
        level = grid.levels[0]
        stime = grid.stime_str
        year = stime[0:4]
        month = stime[4:6]
        day = stime[6:8]
        hour = stime[8:10]
        hour_range = str(grid.dtimes[0])
        values = da.values
        grid_values = np.squeeze(values)
        vmax = math.ceil(max(grid_values.flatten()))
        vmin = math.ceil(min(grid_values.flatten()))

        dif = (vmax - vmin) / 10.0
        if dif ==0:
            inte = 1
        else:
            inte = math.pow(10, math.floor(math.log10(dif)))
        # 用基本间隔，将最大最小值除于间隔后小数点部分去除，最后把间隔也整数化
        r = dif / inte
        if r < 3 and r >= 1.5:
            inte = inte * 2
        elif r < 4.5 and r >= 3:
            inte = inte * 4
        elif r < 5.5 and r >= 4.5:
            inte = inte * 5
        elif r < 7 and r >= 5.5:
            inte = inte * 6
        elif r >= 7:
            inte = inte * 8
        vmin = inte * ((int)(vmin / inte) - 1)
        vmax = inte * ((int)(vmax / inte) + 1)

        end = len(save_path)
        start = max(0, end - 16)

        if title is None:
            title = ("diamond 4 " + save_path[start:end] + "\n"
                     + year + " " + month + " " + day + " " + hour + " " + hour_range + " " + str(level) + "\n"
                     + "{:.6f}".format(grid.dlon) + " " + "{:.6f}".format(grid.dlat) + " " + "{:.6f}".format(grid.slon) + " " +"{:.6f}".format(grid.elon) + " "
                     + "{:.6f}".format(grid.slat) + " " + "{:.6f}".format(grid.elat) + " " + str(grid.nlon) + " " + str(grid.nlat) + " "
                     + str(inte) + " " + str(vmin) + " " + str(vmax) + " 1 0")
        else:

            title = ("diamond 4 "+ title +"\n"
                     + year + " " + month + " " + day + " " + hour + " " + hour_range + " " + str(level) + "\n"
                     + "{:.6f}".format(grid.dlon) + " " + "{:.6f}".format(grid.dlat) + " " + "{:.6f}".format(grid.slon) + " " + "{:.6f}".format(grid.elon) + " "
                     + "{:.6f}".format(grid.slat) + " " + "{:.6f}".format(grid.elat) + " " + str(grid.nlon) + " " + str(grid.nlat) + " "
                     + str(inte) + " " + str(vmin) + " " + str(vmax) + " 1 0")

        # 二维数组写入micaps文件
        format_str = "%." + str(effectiveNum) + "f "

        np.savetxt(save_path, grid_values, delimiter=' ',
                   fmt=format_str, header=title, comments='')
        if show:
            print('成功输出至'+ save_path)
        return True
    except:
        exstr = traceback.format_exc()
        print(exstr)
        return False

def write_griddata_to_nc(da,save_path = "a.txt",creat_dir = False,effectiveNum = 3,show = False):
    try:
        dir = os.path.split(os.path.abspath(save_path))[0]
        if not os.path.isdir(dir):
            if not creat_dir:
                print("文件夹："+dir+"不存在")
                return False
            else:
                meteva.base.tool.path_tools.creat_path(save_path)
        scale_factor = math.pow(10,-effectiveNum)
        #print(scale_factor)
        encodingdict = {da.name:{
                            'dtype': 'int32',
                            'scale_factor': scale_factor,
                             'zlib': True,
                            '_FillValue':None
                            }
                        }

        da.to_netcdf(save_path,encoding = encodingdict)
        if show:
            print('成功输出至' + save_path)
        return True
    except:
        exstr = traceback.format_exc()
        print(exstr)
        return False
def write_griddata_to_micaps11(wind,save_path = "a.txt",creat_dir = False,effectiveNum = 3,show = False,title = None):
    try:
        dir = os.path.split(os.path.abspath(save_path))[0]
        if not os.path.isdir(dir):
            if not creat_dir:
                print("文件夹："+dir+"不存在")
                return False
            else:
                meteva.base.tool.path_tools.creat_path(save_path)
        grid0 = meteva.base.basicdata.get_grid_of_data(wind)
        nlon = grid0.nlon
        nlat = grid0.nlat
        slon = grid0.slon
        slat = grid0.slat
        elon = grid0.elon
        elat = grid0.elat
        dlon = grid0.dlon
        dlat = grid0.dlat
        level = grid0.levels[0]
        stime = grid0.stime_str
        year = stime[0:4]
        month = stime[4:6]
        day = stime[6:8]
        hour = stime[8:10]
        values = wind.values
        grid_values = np.squeeze(values).reshape(2*nlat,nlon)

        end = len(save_path)
        start = max(0, end - 16)

        if title is None:
            title = ("diamond 11 " + save_path[start:end] + "\n"
                     + year + " " + month + " " + day + " " + hour + " " + str(level) + "\n"
                     + str(grid0.dlon) + " " + str(grid0.dlat) + " " + str(grid0.slon) + " " + str(grid0.elon) + " "
                     + str(grid0.slat) + " " + str(grid0.elat) + " " + str(grid0.nlon) + " " + str(grid0.nlat))
        else:

            title  = ("diamond 11 " + title+ "\n"
                    +year + " "+ month + " " + day+ " " +hour +" " + str(level)+"\n"
                    + str(grid0.dlon) + " " + str(grid0.dlat) + " " + str(grid0.slon) + " " + str(grid0.elon) + " "
                    + str(grid0.slat) + " " + str(grid0.elat) + " " + str(grid0.nlon) + " " + str(grid0.nlat))

        format_str = "%." + str(effectiveNum) + "f "

        np.savetxt(save_path, grid_values, delimiter=' ',
                   fmt=format_str, header=title, comments='')
        if show:
            print('成功输出至' + save_path)
        return True
    except:
        exstr = traceback.format_exc()
        print(exstr)
        return False

def tran_griddata_to_gds_flow(da):
    grid0 = meteva.base.get_grid_of_data(da)
    discriminator = b"mdfs"
    if len(grid0.members)==1:
        data_type = 4
    elif len(grid0.members)==2:
        data_type =11
    else:
        print("仅支持micap4类和micaps11类数据输出成GDS格式")
    data_type_byte = np.ndarray.tobytes(np.array([data_type]).astype(np.int16))
    mName = grid0.members[0]
    mName = mName.encode(encoding='utf-8')
    if len(mName)<20:
        mName = mName + np.ndarray.tobytes(np.zeros(20-len(mName)).astype(np.int8))

    eleName = b""
    if "eleName" in da.attrs.keys():
        eleName = da.attrs["eleName"]
    if len(eleName) < 50:
        eleName = eleName + np.ndarray.tobytes(np.zeros(50 - len(eleName)).astype(np.int8))

    description =b""
    if "description" in da.attrs.keys():
        description = da.attrs["description"]
    if len(description) < 30:
        description = description + np.ndarray.tobytes(np.zeros(30 - len(description)).astype(np.int8))


    level = np.ndarray.tobytes(np.array(grid0.levels[0]).astype(np.float32))
    y_m_d_h_timezone_peroid = np.ndarray.tobytes(np.array([2021,1,1,8,8,0]).astype(np.int32))
    slon_elon_dlon = np.ndarray.tobytes(np.array([grid0.slon,grid0.elon,grid0.dlon]).astype(np.float32))
    nlon = np.ndarray.tobytes(np.array([grid0.nlon]).astype(np.int32))
    slat_elat_dlat = np.ndarray.tobytes(np.array([grid0.slat, grid0.elat, grid0.dlat]).astype(np.float32))
    nlat  =  np.ndarray.tobytes(np.array([grid0.nlat]).astype(np.int32))
    vmin,vmax,inte = meteva.base.tool.plot_tools.get_isoline_set(da)
    sValue_eValue_dValue = np.ndarray.tobytes(np.array([vmin,vmax,inte]).astype(np.float32))
    blank = np.ndarray.tobytes(np.zeros(100).astype(np.int8))
    value_bytes = np.ndarray.tobytes(da.values.astype(np.float32))
    bytes1 = discriminator+data_type_byte+mName+eleName+description+level+y_m_d_h_timezone_peroid
    bytes2 = slon_elon_dlon+nlon+slat_elat_dlat+nlat+sValue_eValue_dValue+blank+value_bytes
    bytes = bytes1+bytes2

    return bytes

def write_griddata_to_gds_file(da,save_path = "a.txt",creat_dir = False,show = False):
    try:
        dir = os.path.split(os.path.abspath(save_path))[0]
        if not os.path.isdir(dir):
            if not creat_dir:
                print("文件夹："+dir+"不存在")
                return False
            else:
                meteva.base.tool.path_tools.creat_path(save_path)

        bytes = tran_griddata_to_gds_flow(da)
        br = open(save_path, 'wb')
        br.write(bytes)
        br.close()
        if show:
            print('成功输出至' + save_path)
        return True
    except:
        exstr = traceback.format_exc()
        print(exstr)
        return False

if __name__ == "__main__":
    grd = meteva.base.read_griddata_from_micaps4(r"H:\test_data\input\meb\m4.txt")
    write_griddata_to_gds_file(grd,save_path=r"H:\test_data\output\meb\gds_test.000")


import numpy as np


def rigid_transform(theta, p0, n, cen):
    theta = np.array(theta)
    p0 = np.array(p0)
    cen = np.array(cen)
    if n is None:
        n = p0.shape[0]
    p0cen = p0 - cen
    r = theta[2]
    p0cen = np.hstack((np.cos(r) * p0cen[:, 0] - np.sin(r) * p0cen[:, 1], np.sin(r) * p0cen[:, 0] + np.cos(r) * p0cen[:, 1]))
    res = p0cen.reshape((n, 2), order="F") + np.tile(np.array(theta[0:2]), n).reshape(n, 2, order='C') + cen
    return res



import copy
import pandas as pd
import numpy as np
from .centmatch import centmatch
from .deltamm import deltamm
from .minboundmatch import minboundmatch


def reset_label_id_by_area(look_merge):
    xfeats =look_merge['grd_ob_features']
    yfeats = look_merge['grd_fo_features']
    match_count = look_merge["match_count"]
    area_list = []
    for i in range(match_count):
        area_list.append(xfeats[i+1][0].size)
    area_array = np.array(area_list)
    sort_index = match_count - np.argsort(np.argsort(area_array))

    dlon_dlat = look_merge["grid"].dlon *  look_merge["grid"].dlat
    look_merge_new = copy.deepcopy(look_merge)
    for i in range(match_count):
        j = sort_index[i]
        look_merge_new["grd_ob_label"].values[look_merge["grd_ob_label"].values == i+1] = j
        look_merge_new["grd_fo_label"].values[look_merge["grd_fo_label"].values == i+1] = j
        look_merge_new["grd_ob_features"][j] = look_merge["grd_ob_features"][i+1]
        look_merge_new["grd_fo_features"][j] = look_merge["grd_fo_features"][i+1]

    for j in range(look_merge_new["grd_ob_features"]["label_count"]):
        look_merge_new["grd_ob_features"]["area"].append(look_merge_new["grd_ob_features"][j+1][0].size * dlon_dlat)
    for j in range(look_merge_new["grd_fo_features"]["label_count"]):
        look_merge_new["grd_fo_features"]["area"].append(look_merge_new["grd_fo_features"][j+1][0].size * dlon_dlat)

    return look_merge_new

def merge_force(look_match, verbose=False):
    x = copy.deepcopy(look_match)
    out = {}
    if x.__contains__('implicit_merges') and x.__contains__('merges'):
        print('MergeForce: both implicit_merges and merges components found.  Using merges component.')
        m = x['merges']
    elif not x.__contains__('implicit_merges') and not x.__contains__('merges'):
        if verbose:
            print('\nNo merges found.  Returning object x unaltered.\n')
        return x
    elif not x.__contains__('merges'):
        m = x['implicit_merges']
    else:
        m = x['merges']
    grid0 = x["grid"]
    out['grd_ob'] = x['grd_ob']
    out['grd_fo'] = x['grd_fo']
    out["grd_ob_smooth"] = x["grd_ob_smooth"]
    out["grd_fo_smooth"] = x["grd_fo_smooth"]
    #out['identifier_function'] = x['identifier_function']
    out['identifier_label'] = x['identifier_label']
    out['match_type'] = x['match_type']
    out['match_message'] = ''.join((x['match_message'], " (merged) "))
    xdim = x['grd_ob'].values.squeeze().shape
    #print(xdim)
    if m == None:
        nmatches = 0
    else:
        nmatches = len(m)
    out["match_count"] = nmatches

    ar = np.arange(1, nmatches + 1, 1)
    matches = np.vstack((ar, ar)).T
    matches = pd.DataFrame(matches, columns = ["ob", "fo"])
    out['matches'] = matches.values
    #xp = x['Xlabelsfeature']
    #yp = x['Ylabelsfeature']
    xp = x['grd_ob_features']
    yp = x['grd_fo_features']
    #if len(xp.keys())==0 or len(yp.keys()) ==0:
    if xp["label_count"] == 0 or yp["label_count"] ==0:
        out['grd_ob_features'] = x["grd_ob_features"]
        out['grd_fo_features'] = x["grd_fo_features"]
        out['grd_ob_label'] = x["grd_ob_label"]
        out['grd_fo_label'] = x["grd_fo_label"]
        # vxunmatched = 0
        # if len(xp.keys()) >0:
        #     vxunmatched = list(range(1, len(xp.keys()) + 1))
        # fcunmatched=0
        # if len(yp.keys())>0:
        #     fcunmatched = list(range(1, len(yp.keys()) + 1))
        #out['unmatched'] = {'ob': vxunmatched, 'fo': fcunmatched}
        out["label_list_ob"] = np.arange(1, xp["label_count"] + 1).tolist()
        out["label_list_fo"] = np.arange(1, yp["label_count"] + 1).tolist()
        out["label_list_matched"] = []
        out["unmatched"] ={'ob': np.arange(1, xp["label_count"] + 1).tolist(), 'fo': np.arange(1, yp["label_count"] + 1).tolist()}
        out['MergeForced'] = True
        out["grid"] = copy.deepcopy(look_match["grid"])

    else:

        xfeats = {'Type': xp['Type'], 'xrange': xp['xrange'], 'yrange': xp['yrange'], 'dim': xp['dim'], 'warnings': xp['warnings'],
                  'xstep': xp['xstep'], 'ystep': xp['ystep'], 'area': [], 'xcol': xp['xcol'], 'yrow': xp['yrow']}
        yfeats = {'Type': yp['Type'], 'xrange': yp['xrange'], 'yrange': yp['yrange'], 'dim': yp['dim'], 'warnings': yp['warnings'],
                  'xstep': yp['xstep'], 'ystep': yp['ystep'], 'area': [], 'xcol': yp['xcol'], 'yrow': yp['yrow']}
        xlabeled = np.zeros([xdim[0], xdim[1]], dtype=int)
        ylabeled = np.zeros([xdim[0], xdim[1]], dtype=int)

        if verbose:
            print("Loop through ", nmatches, " components of merges list to set up new (matched) features.\n")
        for i in range(nmatches):
            if verbose:
                print(i, " ")
            tmp = np.array(m[i])-1
            uX = sorted(set(tmp[:, 1]))
            uY = sorted(set(tmp[:, 0]))

            nX = len(uX)
            nY = len(uY)
            xtmp = np.zeros((grid0.nlat,grid0.nlon))
            ytmp = np.zeros((grid0.nlat, grid0.nlon))
            xtmp[xp[uX[0] + 1]] = 1
            ytmp[yp[uY[0] + 1]] = 1

            if nX > 1:
                for j in range(1, nX):
                    xtmp[xp[uX[j] + 1]] = 1

            if nY > 1:
                for k in range(1, nY):
                    ytmp[yp[uY[k] + 1]] = 1

            xfeats[i + 1] = np.where(xtmp>0)
            yfeats[i + 1] = np.where(ytmp>0)
            nozero = np.transpose(np.nonzero(xtmp))
            for j in np.arange(len(nozero)):
                xlabeled[nozero[j][0]][nozero[j][1]] = i + 1
            nozero = np.transpose(np.nonzero(ytmp))
            for j in np.arange(len(nozero)):
                ylabeled[nozero[j][0]][nozero[j][1]] = i + 1
        if x['unmatched']['ob'] == None or x['unmatched']['ob'] == 'NULL':
            unX = x['unmatched']['ob']
            nX2 = 0
        elif type(x['unmatched']['ob']) == int:
            unX = x['unmatched']['ob']-1
            nX2 = 1
        else:
            unX1 = sorted(x['unmatched']['ob'])
            unX = []
            nX2 = len(unX1)
            for nn in range(nX2):
                unX.append(unX1[nn]-1)

        if x['unmatched']['fo'] == None or x['unmatched']['fo'] == 'NULL':
            unY = x['unmatched']['fo']
            nY2 = 0
        elif type(x['unmatched']['fo']) == int:
            unY = x['unmatched']['fo']-1
            nY2 = 1
        else:
            unY1 = sorted(x['unmatched']['fo'])
            nY2 = len(unY1)
            unY = []
            for nn in range(nY2):
                unY.append(unY1[nn]-1)
        if nX2 > 0:
            if verbose:
                print("\nLoop to add/re-label all unmatched observed features.\n")
            vxunmatched = list(range((nmatches + 1), (nmatches + nX2 + 1)))
            for i in range(nX2):
                #xtmp = np.zeros((grid0.nlat,grid0.nlon))
                #xtmp[xp['labels_' + str(unX[i] + 1)]] = 1
                #xfeats[nmatches + i] = xtmp
                xtmp = xp[unX[i] + 1]
                xfeats[i + 1+nmatches] = xtmp
                xlabeled[xtmp] = i + 1+nmatches
                '''
                [rows, cols] = xtmp.shape
                for xi in range(rows):
                    for xj in range(cols):
                        if xtmp[xi, xj]:
                            #xlabeled[xi, xj] = nmatches + i + 1
                            xlabeled[xi, xj] = -1
                            '''
        else:
            vxunmatched = []

        if nY2 > 0:
            if verbose:
                print("\nLoop to add/re-label all unmatched forecast features.\n")
            fcunmatched = list(range((nmatches + 1), (nmatches + nY2 + 1)))

            for i in range(nY2):
                ytmp = yp[unY[i] + 1]
                #ytmp[yp['labels_' + str(unY[i] + 1)]] = 1
                yfeats[i  + 1+nmatches] = ytmp
                #yfeats[nmatches + i] = ytmp
                ylabeled[ytmp] =i + 1+nmatches
                '''
                [rows, cols] = ytmp.shape
                for yi in range(rows):
                    for yj in range(cols):
                        if ytmp[yi, yj]:
                            #ylabeled[yi, yj] = nmatches + i + 1
                            ylabeled[yi, yj] = -1
                            '''

            '''
            for i in range(nY2):
                ytmp = yp['labels_' + str(unY[i] + 1)]
                print(ytmp)
                yfeats['labels_' + str(i + 1)] = ytmp
                #yfeats[nmatches + i] = ytmp
                [rows, cols] = ytmp.shape
                for yi in range(rows):
                    for yj in range(cols):
                        if ytmp[yi, yj]:
                            #ylabeled[yi, yj] = nmatches + i + 1
                            ylabeled[yi, yj] = -1
            '''
        else:
            fcunmatched = []
        xfeats["label_count"] = nmatches +nX2
        yfeats["label_count"] = nmatches +nY2
        out['grd_ob_features'] = xfeats
        out['grd_fo_features'] = yfeats
        grd_ob_labeled = x["grd_ob_label"]
        grd_ob_labeled.values[:] = xlabeled[:]
        out['grd_ob_label'] = grd_ob_labeled
        grd_fo_labeled = x["grd_fo_label"]
        grd_fo_labeled.values[:] = ylabeled[:]
        out['grd_fo_label'] = grd_fo_labeled
        out['unmatched'] = {'ob': vxunmatched, 'fo': fcunmatched}
        out['MergeForced'] = True
        out["grid"] = x["grid"]
        out = reset_label_id_by_area(out)
        out["label_list_ob"]=np.arange(1,out["grd_ob_features"]["label_count"]+1).tolist()
        out["label_list_fo"]=np.arange(1,out["grd_fo_features"]["label_count"]+1).tolist()
        out["label_list_matched"]=np.arange(1,nmatches+1).tolist()

    return out



def merge(look_ff,match_method = centmatch,criteria = 1, const = 14, areafac = 1
          , p = 2, mindist = float('inf'),show = False):
    if match_method == centmatch:
        look_match = centmatch(look_ff,criteria = criteria, const = const,  show = show)
    elif match_method == deltamm:
        look_match = deltamm(look_ff, p = p,show=show)
    else:
        look_match = minboundmatch(look_ff,mindist = mindist, show=show)

    look_merge = merge_force(look_match)
    return look_merge


'''
if __name__ == '__main__':
    data = np.load(r"F:\Work\MODE\tra_test\centmatch\centmatchResult_PA3.npy", allow_pickle=True).tolist()
    look2 = merge_force(data)
    pyplot.imshow(look2['Xlabeled'])
    pyplot.colorbar()
    pyplot.figure(2)
    pyplot.imshow(look2['Ylabeled'])
    pyplot.colorbar()

    levels=np.linspace(0.5,np.max((look2['Xlabeled'], look2['Ylabeled'])) + 0.5,
                       np.max((look2['Xlabeled'], look2['Ylabeled']))+1)
    fig, ax = plt.subplots(1, 2, figsize = (10, 5))
    ax = ax.flatten()
    im1 = ax[0].contourf(look2['Xlabeled'], cmap = 'jet', levels = levels)
    im2 = ax[1].contourf(look2['Ylabeled'], cmap = 'jet', levels = levels)
    fig.colorbar(im2, ax=[ax[0], ax[1]], fraction=0.03, pad=0.05)
    pyplot.show()
    print("hello")
'''






def midpoints_psp(axis, squared=False):
    x = (axis['x1'] + axis['x0']) / 2
    y = (axis['y1'] + axis['y0']) / 2
    #y = (axis['y1'] - axis['y0']) ** 2
    return {'x': x, 'y': y, 'window': axis}



import meteva
import numpy as np

def tran_typhoon_report(sta_all):
    '''
    将台风报文的数据排列方式转换成按观测预报匹配方式排列的数据
    :param sta_all:
    :return:
    '''

    sta0_list,ids = meteva.base.group(sta_all,g = "id")
    sta_list_speed = []
    sta_list_position = []
    sta_value_columns = sta_all.iloc[:, 6:].columns.values.tolist()
    for i in range(len(ids)):
        sta0 = sta0_list[i]
        sta0_ob = meteva.base.sele_by_para(sta0,dtime = 0)
        sta0_ob.sort_values(by = ["time"],inplace = True)
        value = sta0_ob.iloc[:,-1]
        index = np.where(value>=17.2)
        sta1_ob = sta0_ob.iloc[index[0][0]:,:]
        sta1_fo = meteva.base.sele_by_para(sta0,dtime_range = [1,720])
        sta_all_s = meteva.base.combine_on_obTime_id(sta1_ob,[sta1_fo],need_match_ob=True)
        sta_list_speed.append(sta_all_s)
        sta1_ob_p = sta1_ob.drop(sta_value_columns, axis=1, inplace=False)
        sta1_ob_p["x"] = sta1_ob_p["lon"]
        sta1_ob_p["y"] = sta1_ob_p["lat"]
        sta1_fo_p = sta1_fo.drop(sta_value_columns, axis=1, inplace=False)
        sta1_fo_p["x"] = sta1_fo_p["lon"]
        sta1_fo_p["y"] = sta1_fo_p["lat"]
        sta_all_p =  meteva.base.combine_on_obTime_id(sta1_ob_p,[sta1_fo_p],need_match_ob=True)
        sta_list_position.append(sta_all_p)

    sta_all_speed = meteva.base.concat(sta_list_speed)
    meteva.base.set_stadata_names(sta_all_speed,data_name_list=["obs","fst"])
    sta_all_position = meteva.base.concat(sta_list_position)
    meteva.base.set_stadata_names(sta_all_position, data_name_list=["lon_obs", "lat_obs","lon_fo","lat_fo"])

    return sta_all_speed,sta_all_position




if __name__ == "__main__":


    filename = r"H:\test_data\input\meb\cyclone\babj_2201.dat"  # 设置台风路径文件
    sta1 = meteva.base.read_cyclone_trace(filename, id_cyclone=2201, column=meteva.base.m7_element_column.最大风速)  # 读取台风路径数据（包括定位和预报）
    sta_all_speed,sta_all_position = tran_typhoon_report(sta1)

    #result = meteva.product.score(sta_all_speed,meteva.method.rmse,g = "dtime")
    sta_part = meteva.base.sele_by_para(sta_all_position,time = "2022040808")
    print(sta_part)
    result = meteva.product.score(sta_part,meteva.method.distance,on_earth_surface= True,g = "dtime")
    print(result)

